import click
from virl.cli.license.features.show.commands import show as showc
from virl.cli.license.features.update.commands import update as updatec


@click.group()
def features():
    """
    work with licensed features
    """
    pass


features.add_command(showc, name="show")
features.add_command(updatec, name="update")



from collections import OrderedDict
import yaml
from jinja2 import Environment, PackageLoader


def render_topl_template(devices):
    """
    renders topology: section of testbed yaml
    """
    pass


def setup_yaml():
    """ https://stackoverflow.com/a/8661021 """
    represent_dict_order = lambda self, data: self.represent_mapping("tag:yaml.org,2002:map", data.items())  # noqa
    yaml.add_representer(OrderedDict, represent_dict_order)


setup_yaml()


def render_testbed_template(name, servers, devices, topo_devs, conn_class="unicon.Unicon"):
    """
    we need to merge information from multiple sources to generate all
    the required parameters for the topology key in the testbed yaml config

    """
    # sim_name should be the only key in the dictionary
    if len(topo_devs.keys()) != 1:
        raise ValueError("too many keys in topo data")

    sim_name = list(topo_devs)[0]

    try:
        topology = topo_devs[sim_name]
    except KeyError:
        raise Exception("something went wrong")

    j2_env = Environment(loader=PackageLoader("virl"), trim_blocks=False)
    template = j2_env.get_template("pyats/testbed_yaml.j2")
    return template.render(name=sim_name, conn_class=conn_class, servers=servers, devices=devices, topology=topology)


def pyats_testbed_generator1(env, virl_data, roster, interfaces, dev_username="cisco", dev_password="cisco", conn_class="unicon.Unicon"):
    """
    given a sim roster produces a testbed file suitable for
    use with pyats
    """

    # generate a device dictionary for each device
    name = env + "_testbed"

    devices = dict()
    servers = dict()
    topology = interfaces

    for device, props in roster.items():
        # gather information about the device
        virl_server = str(props.get("SimulationHost", None))
        device_type = str(props.get("NodeSubtype", None))
        device_name = str(props.get("NodeName", None))
        console_port = str(props.get("PortConsole", None))

        # map node types to known abstraction
        os = None
        if "NX" in device_type:
            os = "nxos"
        elif "XR" in device_type:
            os = "iosxr"
        elif "CSR" in device_type:
            os = "iosxe"
        elif "IOS" in device_type:
            os = "ios"
        elif "IOS" in device_type:
            os = "ios"
        elif "ASAv" in device_type:
            os = "asa"

        # we prefer external addr
        external_ip = str(props.get("externalAddr", ""))
        mgmt_ip = str(props.get("managementIP", ""))
        if len(external_ip) > 6:
            address = external_ip
        elif len(mgmt_ip) > 6:
            address = mgmt_ip

        # add server devices to testbed->server section
        if device_type in ["mgmt-lxc", "server", "lxc"]:
            servers[device_name] = dict()
            servers[device_name]["server"] = device_name
            servers[device_name]["address"] = address
            servers[device_name]["path"] = ""
            continue

        # some devices can't conform
        elif device_type in ["LXC FLAT"]:
            continue
        # all other devices
        else:
            if device_name == "None":
                continue
            devices[device_name] = dict()
            devices[device_name]["alias"] = device_name
            if os:
                devices[device_name]["os"] = os

            devices[device_name]["type"] = device_type
            devices[device_name]["platform"] = device_type
            devices[device_name]["tacacs"] = {"username": dev_username}
            devices[device_name]["passwords"] = {"tacacs": dev_password, "enable": dev_password, "line": dev_password}
            devices[device_name]["connections"] = OrderedDict()

            devices[device_name]["connections"]["console"] = OrderedDict()
            console_info = {"protocol": "telnet", "ip": virl_server, "port": console_port}
            devices[device_name]["connections"]["console"] = console_info

    topology_yaml = render_testbed_template(name, servers, devices, topology, conn_class=conn_class)

    return topology_yaml


def pyats_testbed_generator(lab):
    return lab.get_pyats_testbed()



from . import BaseTest
from click.testing import CliRunner
from virl.cli.main import virl
try:
    from unittest.mock import patch
except ImportError:
    from mock import patch


class Tests(BaseTest):

    @patch("virl.cli.viz.commands.subprocess.Popen", auto_spec=False)
    def test_virl_uwm(self, call_mock):
        runner = CliRunner()
        runner.invoke(virl, ["viz"])
        url = 'http://localhost:19402/?sim_id=TEST_ENV#/layer/phy'
        call_mock.assert_called_with(['open', url])



from . import BaseTest
from .mocks.api import MockVIRLServer
from click.testing import CliRunner
import requests_mock
from virl.cli.main import virl


class LsTest(BaseTest):

    def test_virl_ls_alL(self):
        with requests_mock.mock() as m:
            # Mock the request to return what we expect from the API.
            m.get('http://localhost:19399/simengine/rest/list',
                  json=MockVIRLServer.list_simulations())
            runner = CliRunner()
            result = runner.invoke(virl, ["ls", "--all"])
            self.assertEqual(0, result.exit_code)

    def test_virl_ls(self):
        with requests_mock.mock() as m:
            # Mock the request to return what we expect from the API.
            m.get('http://localhost:19399/simengine/rest/list',
                  json=MockVIRLServer.list_simulations())
            runner = CliRunner()
            result = runner.invoke(virl, ["ls"])
            self.assertEqual(0, result.exit_code)



# -*- coding: utf-8 -*-
import click
import tabulate


def sim_list_table(node_dict):
    click.secho("Running Simulations", fg="green")
    table = list()
    headers = ["Simulation", "Status", "Launched", "Expires"]
    for node, props in node_dict.items():
        tr = list()
        tr.append(node)
        status = props['status']
        if status == 'ACTIVE':
            color = 'green'
        elif status == 'BUILDING':
            color = 'yellow'
        else:
            color = 'red'
        tr.append(click.style(status, fg=color))
        tr.append(props['launched'])
        tr.append(props.get('expires', 'N/A'))
        table.append(tr)
    # wrap the output in this try/except block as some terminals
    # may have problem with the 'fancy_grid'
    try:
        click.echo(tabulate.tabulate(table, headers, tablefmt="fancy_grid"))
    except UnicodeEncodeError:
        click.echo(tabulate.tabulate(table, headers, tablefmt="grid"))



import click
from virl.api import VIRLServer
from virl import helpers
from virl.helpers import get_cml_client, get_current_lab, safe_join_existing_lab
from virl.generators import pyats_testbed_generator1, pyats_testbed_generator


@click.command()
@click.option("--output", "-o", help="output File name")
def pyats(**kwargs):
    """
    generates a pyats testbed config for a lab
    """
    server = VIRLServer()
    client = get_cml_client(server)

    current_lab = get_current_lab()
    if current_lab:
        lab = safe_join_existing_lab(current_lab, client)
        if lab:
            if kwargs.get("output"):
                # user specified output filename
                file_name = kwargs.get("output")
            else:
                # writes to <lab.id>_testbed.yaml by default
                file_name = "{}_testbed.yaml".format(lab.id)

            testbed = pyats_testbed_generator(lab)

            if testbed:
                click.secho("Writing {}".format(file_name))
                with open(file_name, "w") as fd:
                    fd.write(testbed)
            else:
                click.secho("Failed to get testbed data", fg="red")
                exit(1)
        else:
            click.secho("Failed to find running lab {}".format(current_lab), fg="red")
            exit(1)
    else:
        click.secho("Current lab is not set", fg="red")
        exit(1)


@click.command()
@click.argument("env", default="default")
@click.option("--output", "-o", help="output File name ")
def pyats1(env, **kwargs):
    """
    Generates a pyats testbed config for an environment
    """
    if kwargs.get("output"):
        # user specified output filename
        file_name = kwargs.get("output")
    else:
        # writes to <env>_testbed.yaml by default
        file_name = "{}_testbed.yaml".format(env)

    running = helpers.check_sim_running(env)
    if running:
        sim_name = running
        server = VIRLServer()
        roster = server.get_sim_roster(sim_name)
        # sim_name = "topology-fpyHFs"
        virl_data = server.export(sim_name, ip=True).content
        interfaces = server.get_interfaces(sim_name).json()
        testbed_yaml = pyats_testbed_generator1(sim_name, virl_data, roster=roster, interfaces=interfaces)

        click.secho("Writing {}".format(file_name))
        with open(file_name, "w") as yaml_file:
            yaml_file.write(testbed_yaml)

    else:
        click.secho("couldnt generate testbed for for env: {}".format(env), fg="red")



import os
from . import BaseTest
from virl.api import VIRLServer
try:
    from unittest.mock import patch
except ImportError:
    from mock import patch


class Tests(BaseTest):

    def setUp(self):
        if 'VIRL_HOST' in os.environ:
            del os.environ['VIRL_HOST']
        if 'VIRL_USERNAME' in os.environ:
            del os.environ['VIRL_USERNAME']
        if 'VIRL_PASSWORD' in os.environ:
            del os.environ['VIRL_PASSWORD']

        try:
            os.remove('.virl/default/id')
        except OSError:
            pass

    @patch('virl.api.credentials.get_prop', return_value='fake')
    def test_create_virl_server(self, mock_prop):
        server = VIRLServer()
        self.assertIsInstance(server, VIRLServer)

    @patch('virl.api.credentials._get_from_user', return_value='userinput')
    @patch('virl.api.credentials._get_password', return_value='passwordinput')
    def test_prompt_user(self, user_input, password_input):
        server = VIRLServer()
        user = server.user
        self.assertEqual(user, 'userinput')



import click
import tabulate


def license_details_table(license):
    click.secho("Registration Details")
    print_registration(license["registration"])
    click.secho("Authorization Details")
    print_authorization(license["authorization"])
    click.secho("Features")
    print_features(license["features"])


def license_features_table(license):
    table = list()
    headers = ["ID", "Name", "In Use"]
    for feature in license:
        tr = list()
        tr.append(feature["id"])
        tr.append(feature["name"])
        tr.append(feature["in_use"])

        table.append(tr)

    # wrap the output in this try/except block as some terminals
    # may have problem with the 'fancy_grid'
    try:
        click.echo(tabulate.tabulate(table, headers, tablefmt="fancy_grid"))
    except UnicodeEncodeError:
        click.echo(tabulate.tabulate(table, headers, tablefmt="grid"))


def print_registration(reg_obj):
    table = list()
    headers = ["Status", "Expires", "Smart Account", "Virtual Account", "Registration Time", "Registration Status", "Next Renewal Time"]
    tr = list()
    stat_color = None
    if reg_obj["status"] == "COMPLETED":
        stat_color = "green"
    elif reg_obj["status"] == "IN_PROGRESS":
        stat_color = "yellow"
    else:
        stat_color = "red"
    tr.append(click.style(reg_obj["status"], fg=stat_color))
    tr.append(reg_obj["expires"])
    tr.append(reg_obj["smart_account"])
    tr.append(reg_obj["virtual_account"])
    if reg_obj["register_time"]["attempted"]:
        reg_color = None
        tr.append(reg_obj["register_time"]["attempted"])
        if reg_obj["register_time"]["success"] == "SUCCESS":
            reg_color = "green"
        else:
            reg_color = "red"
        tr.append(click.style(reg_obj["register_time"]["success"], fg=reg_color))
    else:
        tr.append("N/A")
        tr.append("N/A")
    if reg_obj["renew_time"]["scheduled"]:
        tr.append(reg_obj["renew_time"]["scheduled"])
    else:
        tr.append("N/A")

    table.append(tr)

    # wrap the output in this try/except block as some terminals
    # may have problem with the 'fancy_grid'
    try:
        click.echo(tabulate.tabulate(table, headers, tablefmt="fancy_grid"))
    except UnicodeEncodeError:
        click.echo(tabulate.tabulate(table, headers, tablefmt="grid"))


def print_authorization(auth_obj):
    table = list()
    headers = ["Status", "Expires", "Renewal Time", "Renewal Status", "Next Renewal Time"]
    tr = list()
    auth_color = None
    if auth_obj["status"] == "IN_COMPLIANCE":
        auth_color = "green"
    else:
        auth_color = "red"
    tr.append(click.style(auth_obj["status"], fg=auth_color))
    tr.append(auth_obj["expires"])
    if auth_obj["renew_time"]["attempted"]:
        renew_color = None
        if auth_obj["renew_time"]["status"] == "SUCCEEDED":
            renew_color = "green"
        elif auth_obj["renew_time"]["status"] == "NOT STARTED":
            renew_color = "yellow"
        else:
            renew_color = "red"
        tr.append(auth_obj["renew_time"]["attempted"])
        tr.append(click.style(auth_obj["renew_time"]["status"], fg=renew_color))
    else:
        tr.append("N/A")
        tr.append("N/A")
    if auth_obj["renew_time"]["scheduled"]:
        tr.append(auth_obj["renew_time"]["scheduled"])
    else:
        tr.append("N/A")

    table.append(tr)

    # wrap the output in this try/except block as some terminals
    # may have problem with the 'fancy_grid'
    try:
        click.echo(tabulate.tabulate(table, headers, tablefmt="fancy_grid"))
    except UnicodeEncodeError:
        click.echo(tabulate.tabulate(table, headers, tablefmt="grid"))


def print_features(feature_obj):
    table = list()
    headers = ["Name", "Description", "In Use", "Status", "Version"]
    for feature in feature_obj:
        tr = list()
        tr.append(feature["name"])
        tr.append(feature["description"])
        tr.append(feature["in_use"])
        color = None
        if feature["status"] == "IN_COMPLIANCE":
            color = "green"
        elif feature["status"] != "INIT":
            color = "red"
        tr.append(click.style(feature["status"], fg=color))
        tr.append(feature["version"])

        table.append(tr)

    # wrap the output in this try/except block as some terminals
    # may have problem with the 'fancy_grid'
    try:
        click.echo(tabulate.tabulate(table, headers, tablefmt="fancy_grid"))
    except UnicodeEncodeError:
        click.echo(tabulate.tabulate(table, headers, tablefmt="grid"))



import tabulate
import click


def node_list_table(nodes):
    click.secho("Here is a list of nodes in this lab")
    table = list()
    headers = ["ID", "Label", "Type", "State", "Wiped?", "L3 Address(es)"]
    skip_types = []
    for node in nodes:
        tr = list()
        if node.node_definition in skip_types:
            continue

        tr.append(node.id)
        tr.append(node.label)
        tr.append(node.node_definition)

        color = "red"
        if node.is_booted():
            color = "green"
        elif node.is_active():
            color = "yellow"

        tr.append(click.style(node.state, fg=color))
        tr.append(node.state == "DEFINED_ON_CORE")
        intfs = []
        if node.is_booted():
            for i in node.interfaces():
                if i.discovered_ipv4:
                    intfs += i.discovered_ipv4
                if i.discovered_ipv6:
                    intfs += i.discovered_ipv6

        tr.append(",".join(intfs))
        table.append(tr)
    # wrap the output in this try/except block as some terminals
    # may have problem with the 'fancy_grid'
    try:
        click.echo(tabulate.tabulate(table, headers, tablefmt="fancy_grid"))
    except UnicodeEncodeError:
        click.echo(tabulate.tabulate(table, headers, tablefmt="grid"))


def node_list_table1(node_dict):
    click.secho("Here is a list of all the running nodes")
    table = list()
    headers = ["Node", "Type", "State", "Reachable", "Protocol", "Management Address", "External Address"]
    skip_subtypes = ["LXC FLAT"]
    for key, props in node_dict.items():
        tr = list()
        # make sure we have minimally useful data otherwise we won't display it
        if all(k in props for k in ["NodeName", "NodeSubtype"]):
            node = props.get("NodeName", "unknown")
            tr.append(node)

            subtype = props.get("NodeSubtype")

            if subtype in skip_subtypes:
                continue
            tr.append(subtype)

            state = props.get("Status", None)

            if state:
                if state in ["ACTIVE"]:
                    color = "green"
                elif state in ["BUILDING"]:
                    color = "yellow"
                else:
                    color = "red"
            else:
                state = "UNKNOWN"
                color = "red"

            reachable = props.get("Annotation", "N/A")
            if reachable:
                if reachable in ["UNREACHABLE"]:
                    r_color = "red"
                elif reachable in ["REACHABLE"]:
                    r_color = "green"
                else:
                    r_color = "red"
            else:
                state = "UNKNOWN"
                r_color = "red"

            tr.append(click.style(state, fg=color))
            tr.append(click.style(reachable, fg=r_color))
            tr.append(props.get("managementProtocol", "N/A"))
            tr.append(props.get("managementIP", "N/A"))
            tr.append(props.get("externalAddr", "N/A"))

            table.append(tr)
    # wrap the output in this try/except block as some terminals
    # may have problem with the 'fancy_grid'
    try:
        click.echo(tabulate.tabulate(table, headers, tablefmt="fancy_grid"))
    except UnicodeEncodeError:
        click.echo(tabulate.tabulate(table, headers, tablefmt="grid"))



import os
from . import BaseTest
from click.testing import CliRunner
from virl.cli.main import virl


class UseTest(BaseTest):
    def test_virl_use(self):
        runner = CliRunner()
        result = runner.invoke(virl, ["use", "TEST_ENV"])
        path = './.virl/default/id'
        file_exists = os.path.isfile(path)
        self.assertTrue(file_exists)
        self.assertEqual(0, result.exit_code)
        with open(path, 'r') as fh:
            data = fh.read()
        contents_correct = 'TEST_ENV' in data
        self.assertTrue(contents_correct)



import click
import requests


def do_pull(repo, fname):
    click.secho("Pulling from {}".format(repo))
    url = "https://raw.githubusercontent.com/"
    url = url + "{}/master/{}".format(repo, fname)
    resp = requests.get(url)
    if resp.ok:
        with open(fname, "w") as fh:
            fh.write(resp.text)
        click.secho("Saved topology as {}".format(fname), fg="green")
    else:
        click.secho("Error pulling {} - repo not found".format(repo), fg="red")
        exit(1)


@click.command()
@click.argument("repo")
def pull(repo, **kwargs):
    """
    pull topology.yaml from repo
    """
    do_pull(repo, "topology.yaml")


@click.command()
@click.argument("repo")
def pull1(repo, **kwargs):
    """
    pull topology.virl from repo
    """
    do_pull(repo, "topology.virl")



import click
from virl.api import VIRLServer
from requests.exceptions import HTTPError


@click.command()
@click.argument('flavor')
def delete(**kwargs):
    """
    delete a specific flavor
    """

    flavor = kwargs.get('flavor')
    server = VIRLServer(port=80)

    # Attempt to delete the flavor
    try:
        r = server.delete_flavor(flavor)
        print("Flavor '{}' deleted.".format(r['name']))
    except IndexError as ierr:
        print(ierr)
    except HTTPError as herr:
        print("\nFailed to delete flavor '{}':".format(flavor))
        print("\n\t{}".format(herr.response.reason))



from . import BaseCMLTest
from click.testing import CliRunner
import requests_mock
import os


class CMLNodesTest(BaseCMLTest):
    @staticmethod
    def get_l3_addresses(req, ctx):
        response = {
            "n0": {"name": "Lab Net", "interfaces": {}},
            "n1": {
                "name": "rtr-1",
                "interfaces": {
                    "52:54:00:1f:27:95": {"id": "i2", "ip4": ["10.1.1.1"], "ip6": ["fc00::1"], "label": "MgmtEth0/RP0/CPU0/0"},
                    "52:54:00:06:b7:7c": {"id": "i3", "ip4": [], "ip6": [], "label": "donotuse1"},
                    "52:54:00:16:ef:1a": {"id": "i4", "ip4": [], "ip6": [], "label": "donotuse2"},
                    "52:54:00:00:73:28": {"id": "i5", "ip4": [], "ip6": [], "label": "GigabitEthernet0/0/0/0"},
                },
            },
        }
        return response

    def test_cml_nodes(self):
        with requests_mock.Mocker() as m:
            # Mock the request to return what we expect from the API.
            self.setup_mocks(m)
            m.get(self.get_api_path("labs/{}/layer3_addresses".format(self.get_test_id())), json=CMLNodesTest.get_l3_addresses)
            virl = self.get_virl()
            runner = CliRunner()
            result = runner.invoke(virl, ["nodes"])
            self.assertEqual(0, result.exit_code)

    def test_cml_nodes_no_lab(self):
        try:
            os.remove(".virl/current_cml_lab")
        except OSError:
            pass

        with requests_mock.Mocker() as m:
            # Mock the request to return what we expect from the API.
            self.setup_mocks(m)
            virl = self.get_virl()
            runner = CliRunner()
            result = runner.invoke(virl, ["nodes"])
            self.assertEqual(1, result.exit_code)
            self.assertIn("No current lab selected", result.output)

    def test_cml_nodes_bogus_lab(self):
        try:
            os.remove(".virl/current_cml_lab")
        except OSError:
            pass

        src_dir = os.path.realpath(".virl")
        with open(".virl/cached_cml_labs/123456", "w") as fd:
            fd.write("lab: bogus\n")

        os.symlink("{}/cached_cml_labs/123456".format(src_dir), "{}/current_cml_lab".format(src_dir))

        with requests_mock.Mocker() as m:
            # Mock the request to return what we expect from the API.
            self.setup_mocks(m)
            virl = self.get_virl()
            runner = CliRunner()
            result = runner.invoke(virl, ["nodes"])
            os.remove(".virl/cached_cml_labs/123456")
            self.assertEqual(1, result.exit_code)
            self.assertIn("Lab 123456 is not running", result.output)



from . import BaseTest
from .mocks.api import MockVIRLServer
from click.testing import CliRunner
import requests_mock
from virl.cli.main import virl


class IdTest(BaseTest):

    def test_virl_id_nosim(self):
        with requests_mock.mock() as m:
            # Mock the request to return what we expect from the API.
            m.get('http://localhost:19399/simengine/rest/list',
                  json=MockVIRLServer.list_simulations())
            runner = CliRunner()
            result = runner.invoke(virl, ["id"])
            self.assertEqual("virlutils-id\n", result.output)



from . import BaseCMLTest
from click.testing import CliRunner


class TestCMLHelp(BaseCMLTest):
    def test_cml_help(self):
        runner = CliRunner()
        virl = self.get_virl()
        result = runner.invoke(virl, ["--help"])
        self.assertEqual(0, result.exit_code)
        for command in [
            "clear",
            "cockpit",
            "command",
            "console",
            "definitions",
            "down",
            "extract",
            "generate",
            "id",
            "license",
            "ls",
            "nodes",
            "pull",
            "rm",
            "save",
            "search",
            "ssh",
            "start",
            "stop",
            "telnet",
            "ui",
            "up",
            "use",
            "version",
            "wipe",
        ]:
            self.assertIn(command, result.output)



import click
from virl.cli.definitions.nodes.ls.commands import ls
from virl.cli.definitions.nodes.nimport.commands import nimport
from virl.cli.definitions.nodes.export.commands import export


@click.group()
def nodes():
    """
    manage node definitions
    """
    pass


nodes.add_command(ls)
nodes.add_command(export)
nodes.add_command(nimport, name="import")



"""
Collection of utility classes to make getting credentials and
configuration easier.
"""
import getpass
import os
from virl.helpers import find_virl


def _get_from_user(prompt):  # pragma: no cover
    """
    Get the input from the user through interactive prompt.
    """
    resp = input(prompt)
    return resp


def _get_password(prompt):
    """
    Get the password from the user through interactive prompt.
    Using this will ensure that the password is not displayed as
    it is typed.
    """
    return getpass.getpass(prompt)


def _get_from_file(virlrc, prop_name):
    if os.path.isfile(virlrc):
        with open(virlrc) as fh:
            config = fh.readlines()

        for line in config:
            if line.startswith(prop_name):
                prop = line.split("=")[1].strip()
                if prop.startswith('"') and prop.endswith('"'):
                    prop = prop[1:-1]
                return prop


def get_prop(prop_name):
    """
    Gets a variable using the following order

    * Check for .virlrc in current directory

    * recurse up directory tree for .virlrc

    * Check environment variables

    * Check ~/.virlrc

    * Prompt user

    """
    # check for .virlrc in current directory
    cwd = os.getcwd()
    virlrc = os.path.join(cwd, ".virlrc")
    prop = _get_from_file(virlrc, prop_name)

    if prop:
        return prop

    # search up directory tree for a .virlrc
    virl_dir = find_virl()
    if virl_dir:
        virlrc = os.path.join(virl_dir, ".virlrc")
        prop = _get_from_file(virlrc, prop_name)

        if prop:
            return prop

    # try environment next
    prop = os.getenv(prop_name, None)
    if prop:
        return prop

    # check for .virlrc in home directory
    path = os.path.expanduser("~")
    virlrc = os.path.join(path, ".virlrc")
    prop = _get_from_file(virlrc, prop_name)

    return prop or None


def get_credentials(rcfile="~/.virlrc"):
    """
    Used to get the VIRL credentials

    * The login credentials are taken in the following order

    * Check for .virlrc in current directory

    * Check environment variables

    * Check ~/.virlrc

    * Prompt user

    """
    # initialize vars
    host = None
    username = None
    password = None
    config = dict()

    host = get_prop("VIRL_HOST")
    username = get_prop("VIRL_USERNAME")
    password = get_prop("VIRL_PASSWORD")

    # some additional configuration that can be set / overriden
    configurable_props = [
        "VIRL_TELNET_COMMAND",
        "VIRL_CONSOLE_COMMAND",
        "VIRL_SSH_COMMAND",
        "VIRL_SSH_USERNAME",
        "CML_CONSOLE_COMMAND",
        "CML2_PLUS",
        "CML_VERIFY_CERT",
        "CML_DEVICE_USERNAME",
        "CML_DEVICE_PASSWORD",
        "CML_DEVICE_ENABLE_PASSWORD",
        "CML_PLUGIN_PATH",
    ]

    for p in configurable_props:
        if get_prop(p):
            config[p] = get_prop(p)

    if not host:  # pragma: no cover
        prompt = "Please enter the IP / hostname of your virl server: "
        host = _get_from_user(prompt)

    if not username:  # pragma: no cover
        username = _get_from_user("Please enter your VIRL username: ")

    if not password:  # pragma: no cover
        password = _get_password("Please enter your password: ")

    if not all([host, username, password]):  # pragma: no cover
        print("Unable to determine CML/VIRL credentials, please see docs")
        exit(1)
    else:
        return (host, username, password, config)



import click
from virl.cli.definitions.images.iimport.image_file.commands import image_file
from virl.cli.definitions.images.iimport.definition.commands import definition


@click.group()
def iimport():
    """
    import images and image definitions
    """
    pass


iimport.add_command(image_file)
iimport.add_command(definition)



class MockCMLServer(object):
    @staticmethod
    def get_lab_tiles(req, ctx):
        response = {
            "lab_tiles": {
                "5eaea5": {
                    "state": "STOPPED",
                    "created": "2020-07-21 09:10:39",
                    "lab_title": "Other Lab",
                    "lab_description": "",
                    "node_count": 2,
                    "link_count": 1,
                    "id": "5eaea5",
                    "owner": "admin",
                    "topology": {
                        "nodes": [
                            {
                                "id": "n0",
                                "label": "nxos9000-0",
                                "x": -450,
                                "y": -50,
                                "node_definition": "nxosv9000",
                                "image_definition": "nxosv9000-9-2-3",
                                "state": "STOPPED",
                                "cpus": None,
                                "cpu_limit": None,
                                "ram": None,
                                "data_volume": None,
                                "boot_disk_size": None,
                                "tags": [],
                            },
                            {
                                "id": "n1",
                                "label": "xr9kv-0",
                                "x": -150,
                                "y": -50,
                                "node_definition": "iosxrv9000",
                                "image_definition": "iosxrv9000-6-6-2",
                                "state": "STOPPED",
                                "cpus": None,
                                "cpu_limit": None,
                                "ram": None,
                                "data_volume": None,
                                "boot_disk_size": None,
                                "tags": [],
                            },
                        ],
                        "links": [{"id": "l0", "node_a": "n0", "node_b": "n1", "state": "STOPPED"}],
                    },
                },
                "5f0d96": {
                    "state": "STARTED",
                    "created": "2020-08-18 22:47:56",
                    "lab_title": "Mock Test",
                    "lab_description": "",
                    "node_count": 2,
                    "link_count": 1,
                    "id": "5f0d96",
                    "owner": "admin",
                    "topology": {
                        "nodes": [
                            {
                                "id": "n0",
                                "label": "Lab Net",
                                "x": -400,
                                "y": 0,
                                "node_definition": "external_connector",
                                "image_definition": None,
                                "state": "BOOTED",
                                "cpus": None,
                                "cpu_limit": None,
                                "ram": None,
                                "data_volume": None,
                                "boot_disk_size": None,
                                "tags": [],
                            },
                            {
                                "id": "n1",
                                "label": "rtr-1",
                                "x": -200,
                                "y": -50,
                                "node_definition": "iosxrv9000",
                                "image_definition": "iosxrv9000-6-6-2",
                                "state": "BOOTED",
                                "cpus": None,
                                "cpu_limit": None,
                                "ram": None,
                                "data_volume": None,
                                "boot_disk_size": None,
                                "tags": [],
                            },
                            {
                                "id": "n2",
                                "label": "rtr-2",
                                "x": -200,
                                "y": -50,
                                "node_definition": "iosxrv9000",
                                "image_definition": "iosxrv9000-6-6-2",
                                "state": "DEFINED_ON_CORE",
                                "cpus": None,
                                "cpu_limit": None,
                                "ram": None,
                                "data_volume": None,
                                "boot_disk_size": None,
                                "tags": [],
                            },
                        ],
                        "links": [{"id": "l0", "node_a": "n1", "node_b": "n0", "state": "STARTED"}],
                    },
                },
            },
            "health": {
                "valid": True,
                "details": {
                    "cluster_1": {
                        "compute_1": {
                            "has_hw_acceleration": True,
                            "enough_cpus": True,
                            "has_refplat_images": True,
                            "lld_connected": True,
                            "valid": True,
                        },
                        "valid": True,
                    }
                },
                "is_licensed": True,
            },
            "system_information": {"version": "2.1.0-b89.f8805559", "ready": True},
            "system_stats": {
                "clusters": {
                    "cluster_1": {
                        "high_level_drivers": {
                            "compute_1": {
                                "cpu": {
                                    "load": [1.37, 1.45, 1.47],
                                    "count": 16,
                                    "percent": 10.84375,
                                    "model": "Intel(R) Xeon(R) CPU E5-2620 v4 @ 2.10GHz",
                                },
                                "memory": {"total": 101098348544, "free": 86657089536, "used": 13702373376},
                                "disk": {"total": 106285760512, "free": 79234523136, "used": 27051237376},
                                "dominfo": {"allocated_cpus": 2, "allocated_memory": 12582912, "total_vms": 5, "running_vms": 1},
                            }
                        }
                    }
                }
            },
            "user_roles": ["USER", "ADMIN"],
        }
        return response

    @staticmethod
    def get_labs(req, ctx):
        response = ["5eaea5", "5f0d96"]
        return response

    @staticmethod
    def download_lab(req, ctx):
        response = """
        lab:
          description: ''
          notes: ''
          timestamp: 1597805276.8213837
          title: Mock Test
          version: 0.0.3
        nodes:
          - id: n0
            label: Lab Net
            node_definition: external_connector
            x: -400
            y: 0
            configuration: bridge0
            tags: []
            interfaces:
              - id: i0
                slot: 0
                label: port
                type: physical
          - id: n1
            label: rtr-1
            node_definition: iosxrv9000
            x: -200
            y: -50
            configuration: |-
              hostname changeme
              username cisco
              group root-lr
              group cisco-support
              password cisco
              !
              username admin
              group root-lr
              group cisco-support
              password admin
              !
              username lab
              group root-lr
              group cisco-support
              password lab
              !
              end
            image_definition: iosxrv9000-6-6-2
            tags: []
            interfaces:
              - id: i0
                label: Loopback0
                type: loopback
              - id: i1
                slot: 0
                label: MgmtEth0/RP0/CPU0/0
                type: physical
              - id: i2
                slot: 1
                label: donotuse1
                type: physical
              - id: i3
                slot: 2
                label: donotuse2
                type: physical
              - id: i4
                slot: 3
                label: GigabitEthernet0/0/0/0
                type: physical
          - id: n2
            label: rtr-2
            node_definition: iosxrv9000
            x: -200
            y: -50
            configuration: |-
              hostname changeme
              username cisco
              group root-lr
              group cisco-support
              password cisco
              !
              username admin
              group root-lr
              group cisco-support
              password admin
              !
              username lab
              group root-lr
              group cisco-support
              password lab
              !
              end
            image_definition: iosxrv9000-6-6-2
            tags: []
            interfaces:
              - id: i0
                label: Loopback0
                type: loopback
              - id: i1
                slot: 0
                label: MgmtEth0/RP0/CPU0/0
                type: physical
              - id: i2
                slot: 1
                label: donotuse1
                type: physical
              - id: i3
                slot: 2
                label: donotuse2
                type: physical
              - id: i4
                slot: 3
                label: GigabitEthernet0/0/0/0
                type: physical
        links:
          - id: l0
            i1: i1
            n1: n1
            i2: i0
            n2: n0
        """
        return response

    @staticmethod
    def download_alt_lab(req, ctx):
        response = """
        lab:
          description: ''
          notes: ''
          timestamp: 1595337039.0416706
          title: Other Lab
          version: 0.0.3
        nodes:
          - id: n0
            label: nxos9000-0
            node_definition: nxosv9000
            x: -450
            y: -50
            configuration: hostname inserthostname_here
            image_definition: nxosv9000-9-2-3
            tags: []
            interfaces:
              - id: i0
                label: Loopback0
                type: loopback
              - id: i1
                slot: 0
                label: mgmt0
                type: physical
              - id: i2
                slot: 1
                label: Ethernet1/1
                type: physical
              - id: i3
                slot: 2
                label: Ethernet1/2
                type: physical
              - id: i4
                slot: 3
                label: Ethernet1/3
                type: physical
          - id: n1
            label: xr9kv-0
            node_definition: iosxrv9000
            x: -150
            y: -50
            configuration: |-
              hostname changeme
              username cisco
              group root-lr
              group cisco-support
              password cisco
              !
              username admin
              group root-lr
              group cisco-support
              password admin
              !
              username lab
              group root-lr
              group cisco-support
              password lab
              !
              end
            image_definition: iosxrv9000-6-6-2
            tags: []
            interfaces:
              - id: i0
                label: Loopback0
                type: loopback
              - id: i1
                slot: 0
                label: MgmtEth0/RP0/CPU0/0
                type: physical
              - id: i2
                slot: 1
                label: donotuse1
                type: physical
              - id: i3
                slot: 2
                label: donotuse2
                type: physical
              - id: i4
                slot: 3
                label: GigabitEthernet0/0/0/0
                type: physical
        links:
          - id: l0
            i1: i1
            n1: n0
            i2: i1
            n2: n1
        """
        return response

    @staticmethod
    def get_sys_info(req, ctx):
        response = {"version": "2.1.0", "ready": True}
        return response

    @staticmethod
    def auth_ok(req, ctx):
        return "OK"

    @staticmethod
    def authenticate(req, ctx):
        return "1234567890"

    @staticmethod
    def print_req(req, ctx):
        s = "!!!URL: {}, method = {}, params = {}".format(req.url, req.method, req.path)
        print(s)
        return s

    @staticmethod
    def get_lab_element_state(req, ctx):
        return MockCMLServer._get_lab_element_state(req, ctx)

    @staticmethod
    def get_lab_element_state_down(req, ctx):
        return MockCMLServer._get_lab_element_state(req, ctx, "STOPPED")

    @staticmethod
    def _get_lab_element_state(req, ctx, n2_state="BOOTED"):
        response = {
            "nodes": {"n0": "BOOTED", "n1": n2_state, "n2": "DEFINED_ON_CORE"},
            "links": {"l0": "STARTED"},
            "interfaces": {"i0": "STARTED", "i1": "STARTED", "i2": "STARTED", "i3": "STARTED", "i4": "STARTED", "i5": "STARTED"},
        }
        return response

    @staticmethod
    def get_topology(req, ctx):
        response = {
            "nodes": [
                {
                    "id": "n0",
                    "data": {
                        "node_definition": "external_connector",
                        "image_definition": None,
                        "label": "Lab Net",
                        "configuration": "bridge0",
                        "x": -400,
                        "y": 0,
                        "state": "BOOTED",
                        "ram": None,
                        "cpus": None,
                        "cpu_limit": None,
                        "data_volume": None,
                        "boot_disk_size": None,
                        "tags": [],
                    },
                },
                {
                    "id": "n1",
                    "data": {
                        "node_definition": "iosxrv9000",
                        "image_definition": "iosxrv9000-6-6-2",
                        "label": "rtr-1",
                        "configuration": "hostname changeme\n",
                        "x": -200,
                        "y": -50,
                        "state": "BOOTED",
                        "ram": None,
                        "cpus": None,
                        "cpu_limit": None,
                        "data_volume": None,
                        "boot_disk_size": None,
                        "tags": [],
                    },
                },
                {
                    "id": "n2",
                    "data": {
                        "node_definition": "iosxrv9000",
                        "image_definition": "iosxrv9000-6-6-2",
                        "label": "rtr-2",
                        "configuration": "hostname changeme\n",
                        "x": -200,
                        "y": -50,
                        "state": "DEFINED_ON_CORE",
                        "ram": None,
                        "cpus": None,
                        "cpu_limit": None,
                        "data_volume": None,
                        "boot_disk_size": None,
                        "tags": [],
                    },
                },
            ],
            "links": [{"id": "l0", "interface_a": "i2", "interface_b": "i0", "data": {"state": "STARTED"}}],
            "interfaces": [
                {"id": "i0", "node": "n0", "data": {"label": "port", "slot": 0, "state": "STARTED", "type": "physical"}},
                {"id": "i1", "node": "n1", "data": {"label": "Loopback0", "slot": None, "state": "STARTED", "type": "loopback"}},
                {"id": "i2", "node": "n1", "data": {"label": "MgmtEth0/RP0/CPU0/0", "slot": 0, "state": "STARTED", "type": "physical"}},
                {"id": "i3", "node": "n1", "data": {"label": "donotuse1", "slot": 1, "state": "STARTED", "type": "physical"}},
                {"id": "i4", "node": "n1", "data": {"label": "donotuse2", "slot": 2, "state": "STARTED", "type": "physical"}},
                {"id": "i5", "node": "n1", "data": {"label": "GigabitEthernet0/0/0/0", "slot": 3, "state": "STARTED", "type": "physical"}},
            ],
            "lab_notes": "",
            "lab_title": "Mock Test",
            "lab_description": "",
            "state": "STARTED",
            "created_timestamp": 1597805276.8213837,
            "cluster_id": "cluster_1",
            "version": "0.0.3",
        }
        return response

    @staticmethod
    def get_alt_topology(req, ctx):
        response = {
            "nodes": [
                {
                    "id": "n0",
                    "data": {
                        "node_definition": "nxosv9000",
                        "image_definition": "nxosv9000-9-2-3",
                        "label": "nxos9000-0",
                        "configuration": "hostname inserthostname_here",
                        "x": -450,
                        "y": -50,
                        "state": "STOPPED",
                        "ram": None,
                        "cpus": None,
                        "cpu_limit": None,
                        "data_volume": None,
                        "boot_disk_size": None,
                        "tags": [],
                    },
                },
                {
                    "id": "n1",
                    "data": {
                        "node_definition": "iosxrv9000",
                        "image_definition": "iosxrv9000-6-6-2",
                        "label": "xr9kv-0",
                        "configuration": "hostname changeme\n",
                        "x": -150,
                        "y": -50,
                        "state": "STOPPED",
                        "ram": None,
                        "cpus": None,
                        "cpu_limit": None,
                        "data_volume": None,
                        "boot_disk_size": None,
                        "tags": [],
                    },
                },
            ],
            "links": [{"id": "l0", "interface_a": "i1", "interface_b": "i6", "data": {"state": "STOPPED"}}],
            "interfaces": [
                {"id": "i0", "node": "n0", "data": {"label": "Loopback0", "slot": None, "state": "STOPPED", "type": "loopback"}},
                {"id": "i1", "node": "n0", "data": {"label": "mgmt0", "slot": 0, "state": "STOPPED", "type": "physical"}},
                {"id": "i2", "node": "n0", "data": {"label": "Ethernet1/1", "slot": 1, "state": "STOPPED", "type": "physical"}},
                {"id": "i3", "node": "n0", "data": {"label": "Ethernet1/2", "slot": 2, "state": "STOPPED", "type": "physical"}},
                {"id": "i4", "node": "n0", "data": {"label": "Ethernet1/3", "slot": 3, "state": "STOPPED", "type": "physical"}},
                {"id": "i5", "node": "n1", "data": {"label": "Loopback0", "slot": None, "state": "STOPPED", "type": "loopback"}},
                {"id": "i6", "node": "n1", "data": {"label": "MgmtEth0/RP0/CPU0/0", "slot": 0, "state": "STOPPED", "type": "physical"}},
                {"id": "i7", "node": "n1", "data": {"label": "donotuse1", "slot": 1, "state": "STOPPED", "type": "physical"}},
                {"id": "i8", "node": "n1", "data": {"label": "donotuse2", "slot": 2, "state": "STOPPED", "type": "physical"}},
                {"id": "i9", "node": "n1", "data": {"label": "GigabitEthernet0/0/0/0", "slot": 3, "state": "STOPPED", "type": "physical"}},
            ],
            "lab_notes": "",
            "lab_title": "Other Lab",
            "lab_description": "",
            "state": "STOPPED",
            "created_timestamp": 1595337039.0416706,
            "cluster_id": "cluster_1",
            "version": "0.0.3",
        }
        return response



from . import BaseTest
from .mocks.github import MockGitHub # noqa
from click.testing import CliRunner
import requests_mock
from virl.cli.main import virl
import os
try:
    from unittest.mock import patch
except ImportError:
    from mock import patch # noqa


class Tests(BaseTest):

    def setUp(self):
        os.environ['VIRL_HOST'] = 'localhost'
        os.environ['VIRL_USERNAME'] = 'guest'
        os.environ['VIRL_PASSWORD'] = 'guest'

        try:
            os.remove('.virl/default/id')
            os.system('cp tests/v1/static/fake_repo_topology.virl topology.virl')

        except OSError:
            pass

    def test_01_virl_up(self):

        try:
            os.remove('.virl/default/id')
        except OSError:
            pass

        with requests_mock.mock() as m:
            # Mock the request to return what we expect from the API.
            up_url = 'http://localhost:19399/simengine/rest/launch'
            m.post(up_url, json=self.mock_up_response())
            net_url = 'http://localhost:19399/openstack/rest/networks'
            m.get(net_url, json=self.mock_os_net_response())
            runner = CliRunner()
            result = runner.invoke(virl, ["up"])
            self.assertEqual(0, result.exit_code)

    @patch("virl.cli.up.commands.call", auto_spec=False)
    def test_02_virl_up_from_repo(self, call_mock):

        try:
            os.remove('.virl/default/id')
            os.remove('topology.virl')
        except OSError:
            pass

        with requests_mock.mock() as m:
            # Mock the request to return what we expect from the API.
            up_url = 'http://localhost:19399/simengine/rest/launch'
            m.post(up_url, json=self.mock_up_response())
            net_url = 'http://localhost:19399/openstack/rest/networks'
            m.get(net_url, json=self.mock_os_net_response())
            topo_url = 'https://raw.githubusercontent.com/'
            topo_url += 'foo/bar/master/topology.virl'
            m.get(topo_url, json=MockGitHub.get_topology())
            runner = CliRunner()
            runner.invoke(virl, ["up", "foo/bar"])

    def mock_up_response(self):
        response = u'TEST_ENV'
        return response

    def mock_os_net_response(self):
        response = [
            {u'Description': u'L3 SNAT External',
             u'ID': u'e565b39a-ee48-4d35-b22f-15ca13cca40d',
             u'DNS': [u'8.8.8.8', u'171.70.168.183'],
             u'CIDR': u'10.10.241.208/28',
             u'Gateway': u'10.10.241.209',
             u'Network Name': u'ext-net'},
            {u'Description': u'L2 FLAT',
             u'ID': u'426a32c0-4964-4288-ba74-7db1f1f0b848',
             u'DNS': [u'8.8.8.8', u'171.70.168.183'],
             u'CIDR': u'10.10.241.224/27',
             u'Gateway': u'10.10.241.225',
             u'Network Name': u'flat'},
            {u'Description': u'L2 FLAT',
             u'ID': u'bfa4c58b-e404-45af-905d-b50aa603df20',
             u'DNS': [u'8.8.8.8', u'171.70.168.183'],
             u'CIDR': u'172.16.1.192/27',
             u'Gateway': u'172.16.1.193',
             u'Network Name': u'flat1'}
        ]
        return response



import requests
from .credentials import get_credentials


class VIRLServer(object):
    def __init__(self, host=None, user=None, passwd=None, port=19399):

        self._host, self._user, self._passwd, self._config = get_credentials()
        self._port = port
        self.base_api = "http://{}:{}".format(self.host, self._port)

    @property
    def host(self):
        return self._host

    @host.setter
    def host(self, host):
        self._host = host

    @property
    def port(self):
        return self._port

    @port.setter
    def port(self, port):
        self._port = port

    @property
    def user(self):
        return self._user

    @user.setter
    def user(self, user):
        self._user = user

    @property
    def passwd(self):
        return self._passwd

    @passwd.setter
    def passwd(self, passwd):
        self._passwd = passwd

    @property
    def _headers(self):
        return {"Content-Type": "text/xml;charset=UTF-8"}

    @property
    def config(self):
        return self._config

    def get(self, url):
        r = requests.get(url, auth=(self.user, self.passwd), headers=self._headers)
        r.raise_for_status()
        return r

    def get_version(self):
        url = self.base_api + "/roster/rest/test"
        r = self.get(url)
        return r.json()

    def post(self, url, data):
        r = requests.post(url, auth=(self.user, self.passwd), headers=self._headers, data=data)
        r.raise_for_status()
        return r

    def put(self, url, data):
        r = requests.put(url, auth=(self.user, self.passwd), headers=self._headers, data=data)
        r.raise_for_status()
        return r

    def post_with_params(self, url, data, params):
        r = requests.post(url, auth=(self.user, self.passwd), headers=self._headers, params=params, data=data)
        r.raise_for_status()
        return r

    def delete(self, url):
        r = requests.delete(url, auth=(self.user, self.passwd), headers=self._headers)
        r.raise_for_status()
        return r

    def list_simulations(self):
        url = self.base_api + "/simengine/rest/list"
        r = self.get(url)
        return r.json()["simulations"]

    def launch_simulation(self, simulation_name, simulation_data):
        u = self.base_api + "/simengine/rest/launch?session={}"
        u = u.format(simulation_name)
        r = self.post(u, simulation_data)
        return r

    def stop_simulation(self, simulation):
        u = self.base_api + "/simengine/rest/stop/{}".format(simulation)
        r = self.get(u)
        return r

    def get_node_summary(self, simulation):
        url = self.base_api + "/simengine/rest/nodes/{}".format(simulation)
        r = requests.get(url, auth=(self.user, self.passwd))
        nodes = r.json()[simulation]
        return nodes

    def get_node_list(self, simulation):
        """
        returns a list of node,status tuples
        """
        nodes = self.get_node_summary(simulation)
        return nodes.keys()

    def check_node_reachable(self, simulation, node_name):
        """
        returns a list of node,status tuples
        """
        nodes = self.get_node_summary(simulation)
        return nodes[node_name]["reachable"]

    def export(self, simulation, ip=False):
        """
        export simulation to local virl file
        """
        url = self.base_api + "/simengine/rest/export/{}".format(simulation)
        url += "?running-configs=config"
        if ip:
            url += "&updated=true"
        r = self.get(url)
        return r

    def get_node_console(self, simulation, node=None, mode="telnet", port="0"):
        u = self.base_api + "/simengine/rest/serial_port/{}".format(simulation)
        u += "?mode={}&?port={}".format(mode, port)
        if node is not None:
            u += "&nodes={}".format(node)
        r = self.get(u)
        return r

    def get_logs(self, simulation):
        u = self.base_api + "/simengine/rest/events/{}".format(simulation)
        r = self.get(u)
        return r

    def get_sim_roster(self, simulation):
        """
        return a roster entry for given sim
        """
        sim_key = "{}|{}|".format(self.user, simulation)
        u = self.base_api + "/roster/rest/"
        r = self.get(u)
        roster = r.json()
        ret = dict()
        for sim in roster.keys():
            if sim.startswith(sim_key):
                ret[sim] = roster[sim]
        return ret

    def get_interfaces(self, simulation_name):
        """
        returns all interfaces for a simulation
        """
        u = self.base_api + "/simengine/rest/interfaces/{}"
        u = u.format(simulation_name)
        r = self.get(u)
        return r

    def stop_node(self, simulation, node):
        """
        stops a `node` in `simulation`
        """
        u = self.base_api + "/simengine/rest/update/{}/stop?nodes={}"
        u = u.format(simulation, node)
        r = self.put(u, None)
        return r

    def start_node(self, simulation, node):
        """
        stops a `node` in `simulation`
        """
        u = self.base_api + "/simengine/rest/update/{}/start?nodes={}"
        u = u.format(simulation, node)
        r = self.put(u, None)
        return r

    def get_gateway_for_network(self, network):
        u = self.base_api + "/openstack/rest/networks"
        r = self.get(u)
        for net in r.json():
            if net.get("Network Name") == network:
                return net["Gateway"]
        return None

    def get_dns_server_for_network(self, network):
        u = self.base_api + "/openstack/rest/networks"
        r = self.get(u)
        for net in r.json():
            if net.get("Network Name") == network:
                try:
                    return net["DNS"][0]
                except IndexError:
                    return None
        return None

    def get_flavor_id(self, flavor):
        r = self.get_flavors()

        for f in list(r):
            if f["name"] == flavor:
                return f["id"]

        raise IndexError("Flavor {} not found".format(flavor))

    def get_flavors(self, flavor_id=None):
        u = self.base_api + "/rest/flavors"

        if flavor_id is not None:
            u = u + "/{}".format(flavor_id)
        r = self.get(u)

        if flavor_id:
            return r.json()["flavor"]
        return r.json()["flavors"]

    def add_flavor(self, flavor=None, memory=None, vcpus=None):
        u = self.base_api + "/rest/flavors"
        data = {
            "name": flavor,
            "ram": memory,
            "vcpus": vcpus,
        }
        r = self.post_with_params(u, None, data)
        return r.json()["flavor"]

    def delete_flavor(self, flavor):
        f = self.get_flavor_id(flavor)
        u = self.base_api + "/rest/flavors/{}"
        u = u.format(f)
        r = self.delete(u)
        return r.json()["flavor"]



import click
from virl.api import VIRLServer, ViewerPlugin, NoPluginError
from virl.cli.views import node_list_table1, node_list_table
from virl.helpers import get_cml_client, get_current_lab, safe_join_existing_lab
from virl import helpers


@click.command()
def nodes():
    """
    get node list for the current lab
    """
    server = VIRLServer()
    client = get_cml_client(server)

    current_lab = get_current_lab()
    if current_lab:
        lab = safe_join_existing_lab(current_lab, client)
        if lab:
            try:
                pl = ViewerPlugin(viewer="node")
                pl.visualize(nodes=lab.nodes())
            except NoPluginError:
                node_list_table(lab.nodes())
        else:
            click.secho("Lab {} is not running".format(current_lab), fg="red")
            exit(1)
    else:
        click.secho("No current lab selected", fg="red")
        exit(1)


@click.command()
@click.argument("env", default="default")
def nodes1(env, **kwargs):
    """
    get nodes for sim_name
    """
    running = helpers.check_sim_running(env)
    if running:
        sim_name = running
        server = VIRLServer()
        details = server.get_sim_roster(sim_name)
        node_list_table1(details)
    else:
        click.secho("Environment {} is not running".format(env), fg="red")



class MockNSOServer:
    @classmethod
    def launch_simulation(cls):
        response = u"TEST_ENV"
        return response

    @classmethod
    def get_node_console(cls):
        sim_response = {u"router2": u"10.94.241.194:17002", u"router1": u"10.94.241.194:17001"}
        return sim_response

    @classmethod
    def update_devices(cls):
        return {}

    @classmethod
    def perform_sync_from(cls):
        response = {"tailf-ncs:output": {"sync-result": [{"device": "router1", "result": True}, {"device": "router2", "result": True}]}}
        return response

    @classmethod
    def get_module_list(cls):
        response = {
            "ietf-yang-library:module": [
                {
                    "name": "cisco-ios-cli-6.42",
                    "revision": "",
                    "schema": "http://localhost:8080/restconf/tailf/modules/cisco-ios-cli-6.42",
                    "namespace": "http://tail-f.com/ns/ned-id/cisco-ios-cli-6.42",
                    "conformance-type": "import",
                },
            ],
        }
        return response

    @classmethod
    def get_ned_list(cls):
        response = {
            "tailf-ncs:ned-id": [
                {
                    "id": "cisco-ios-cli-6.42:cisco-ios-cli-6.42",
                    "module": [
                        {"name": "ietf-interfaces", "revision": "2014-05-08", "namespace": "urn:ietf:params:xml:ns:yang:ietf-interfaces"},
                        {"name": "ietf-ip", "revision": "2014-06-16", "namespace": "urn:ietf:params:xml:ns:yang:ietf-ip"},
                        {"name": "tailf-ned-cisco-ios", "revision": "2020-01-03", "namespace": "urn:ios"},
                        {"name": "tailf-ned-cisco-ios-id", "namespace": "urn:ios-id"},
                        {"name": "tailf-ned-cisco-ios-stats", "namespace": "urn:ios-stats"},
                    ],
                },
            ],
        }
        return response



import os
import unittest
from click.testing import CliRunner
from .mocks import MockCMLServer
import functools
import sys
import traceback
import pdb
import requests_mock


def setup_cml_environ():
    os.environ["VIRL_HOST"] = "localhost"
    os.environ["VIRL_USERNAME"] = "admin"
    os.environ["VIRL_PASSWORD"] = "admin"
    os.environ["CML2_PLUS"] = "true"
    os.environ["CML_VERIFY_CERT"] = "false"
    os.environ["NSO_HOST"] = "localhost"
    os.environ["NSO_USERNAME"] = "admin"
    os.environ["NSO_PASSWORD"] = "admin"


def debug_on(*exceptions):
    if not exceptions:
        exceptions = (AssertionError,)

    def decorator(f):
        @functools.wraps(f)
        def wrapper(*args, **kwargs):
            try:
                return f(*args, **kwargs)
            except exceptions:
                info = sys.exc_info()
                traceback.print_exception(*info)
                pdb.post_mortem(info[2])

        return wrapper

    return decorator


class BaseCMLTest(unittest.TestCase):
    def setUp(self):
        try:
            os.makedirs(".virl")
        except OSError:
            pass
        # Only doing this because we don't have a better way of controlling
        # injection of VIRL_HOST
        virl = self.get_virl()
        runner = CliRunner()
        with requests_mock.Mocker() as m:
            self.setup_mocks(m)
            runner.invoke(virl, ["use", self.get_test_title()])

    def get_virl(self):
        # This bit of hackery is done since coverage loads all modules into the same
        # namespace.  We need to reload our virl module to get this to recognize
        # the new environment.
        try:
            del sys.modules["virl.cli.main"]
        except KeyError:
            pass

        setup_cml_environ()
        from virl.cli.main import virl

        return virl

    def get_api_path(self, path):
        return "https://localhost/api/v0/{}".format(path)

    def get_test_id(self):
        return "5f0d96"

    def get_test_title(self):
        return "Mock Test"

    def get_alt_id(self):
        return "5eaea5"

    def get_alt_title(self):
        return "Other Lab"

    def setup_mocks(self, m):
        m.get(self.get_api_path("labs"), json=MockCMLServer.get_labs)
        m.get(self.get_api_path("populate_lab_tiles"), json=MockCMLServer.get_lab_tiles)
        m.get(self.get_api_path("labs/{}/download".format(self.get_test_id())), text=MockCMLServer.download_lab)
        m.get(
            self.get_api_path("labs/{}/topology?exclude_configurations=False".format(self.get_test_id())), json=MockCMLServer.get_topology
        )
        m.get(
            self.get_api_path("labs/{}/topology?exclude_configurations=False".format(self.get_alt_id())),
            json=MockCMLServer.get_alt_topology,
        )
        m.get(self.get_api_path("labs/{}/topology?exclude_configurations=True".format(self.get_test_id())), json=MockCMLServer.get_topology)
        m.get(
            self.get_api_path("labs/{}/topology?exclude_configurations=True".format(self.get_alt_id())),
            json=MockCMLServer.get_alt_topology,
        )
        m.get(self.get_api_path("labs/{}/download".format(self.get_alt_id())), text=MockCMLServer.download_alt_lab)
        m.get(self.get_api_path("labs/{}/lab_element_state".format(self.get_test_id())), json=MockCMLServer.get_lab_element_state)
        m.get(self.get_api_path("system_information"), json=MockCMLServer.get_sys_info)
        m.get(self.get_api_path("authok"), text=MockCMLServer.auth_ok)
        m.post(self.get_api_path("authenticate"), text=MockCMLServer.authenticate)
        m.get(self.get_api_path("labs/{}/state".format(self.get_test_id())), json="STARTED")
        m.get(self.get_api_path("labs/{}/state".format(self.get_alt_id())), json="STOPPED")
        m.get(self.get_api_path("labs/{}/check_if_converged".format(self.get_test_id())), json=True)

    def add_debug_mock(self, m):
        m.register_uri(requests_mock.ANY, requests_mock.ANY, text=MockCMLServer.print_req)



import click
import random
import string
import os
import shutil
import errno
import platform
import ctypes
import logging
import ipaddress
import sys
import concurrent.futures
from requests.exceptions import HTTPError
from virl2_client import ClientLibrary


# http://code.activestate.com/recipes/578035-disable-file-system-redirector/
# https://github.com/CiscoDevNet/virlutils/issues/45
class disable_file_system_redirection:
    if platform.system() == "Windows":
        _disable = ctypes.windll.kernel32.Wow64DisableWow64FsRedirection
        _revert = ctypes.windll.kernel32.Wow64RevertWow64FsRedirection

    def __enter__(self):
        self.old_value = ctypes.c_long()
        self.success = self._disable(ctypes.byref(self.old_value))

    def __exit__(self, type, value, traceback):
        if self.success:
            self._revert(self.old_value)


# Taken from https://stackoverflow.com/a/600612/119527
def mkdir_p(path):
    try:
        os.makedirs(path)
    except OSError as exc:  # Python >2.5
        if exc.errno == errno.EEXIST and os.path.isdir(path):
            pass
        else:
            raise  # noqa


def safe_open_w(path):
    """Open "path" for writing, creating any parent directories as needed."""
    mkdir_p(os.path.dirname(path))
    return open(path, "w")


def store_sim_info(name, env="default"):
    with safe_open_w("./.virl/{}/id".format(env)) as f:
        f.write(name)


def remove_sim_info(env="default"):
    path = "./.virl/{}".format(env)
    click.secho("Removing {}".format(path))
    try:
        shutil.rmtree(path)
    except OSError:
        click.secho("Could not remove {}".format(path))


def generate_sim_id():
    letters = string.ascii_letters
    digits = string.digits
    return "".join(random.choice(letters + digits) for _ in range(6))


def get_env_sim_name(env):
    fname = "./.virl/{}/id".format(env)
    with open(fname, "r") as f:
        sim_name = f.read()

    return sim_name


def find_virl():
    pwd = os.getcwd().split(os.sep)
    root = os.path.abspath(os.sep)
    while pwd != root:
        if platform.system() == "Windows":
            lookin = "\\".join(pwd)
            if lookin == "":
                lookin = "\\"
        else:
            lookin = os.path.join(os.sep, *pwd)
        if ".virl" in os.listdir(lookin):
            return lookin
        try:
            pwd.pop()
        except IndexError:
            return None


def check_sim_running(env):
    """
    determines if a sim is already running for a given environment
    """
    try:
        virl_root = find_virl()
        fname = virl_root + "/.virl/{}/id".format(env)
        with open(fname, "r") as f:
            sim_name = f.read()
        if sim_name:
            return sim_name
        else:
            return None
    except Exception:
        return None


def get_mgmt_lxc_ip(sim_roster):
    # grab mgmt-lxc info in case we need it later
    for k, v in sim_roster.items():
        if k.endswith("mgmt-lxc"):
            lxc_ip = v.get("externalAddr", None)
            print("lxc is at {}".format(lxc_ip))
    return lxc_ip


def get_node_from_roster(name, roster):
    for k, v in roster.items():
        if k.endswith(name):
            return v


"""
CML helper functions
"""


def find_virl_or_else():
    """
    if a .virl directory does not exist in our path tree, use the current
    working directory to store it
    """
    virl_root = find_virl()
    if not virl_root:
        return "."

    return virl_root


def get_cache_root():
    """
    get the path to the directory to store cached labs
    """
    virl_root = find_virl_or_else()
    return virl_root + "/.virl/cached_cml_labs"


def get_current_lab_link():
    """
    get the path to the symlink representing the current lab
    """
    virl_root = find_virl_or_else()
    return virl_root + "/.virl/current_cml_lab"


def get_default_plugin_dir():
    """
    get the default plugin directory if one isn't otherwise configured
    """
    virl_root = find_virl_or_else()
    return virl_root + "/.virl/plugins"


def safe_join_existing_lab(lab_id, client):
    """
    gets a lab by its ID only if it exists on the server
    """
    if lab_id in client.get_lab_list():
        return client.join_existing_lab(lab_id)

    return None


def safe_join_existing_lab_by_title(lab_name, client):
    """
    gets a lab ID using its name/title

    This will return None if multiple labs exist for a given name/title.
    Since CML allows duplicate titles, we don't know what the user may want.
    """
    labs = client.find_labs_by_title(lab_name)
    if len(labs) == 1:
        return labs[0]

    return None


def check_lab_cache(lab_id):
    """
    determines if a given lab ID is in the local topology cache
    """
    try:
        cache_root = get_cache_root()
        fname = "{}/{}".format(cache_root, lab_id)
        if os.path.exists(fname):
            return fname
    except Exception:
        pass

    return None


def cache_lab(lab, force=False):
    """
    save a topology YAML file into a local cache
    """
    topo = None
    topo = lab.download()

    cache_lab_data(lab.id, topo, force)


def cache_lab_data(lab_id, lab_data, force=False):
    """
    save lab topology data into a local cache
    """
    cache_root = get_cache_root()
    fname = "{}/{}".format(cache_root, lab_id)
    if force or not os.path.exists(fname):
        with safe_open_w(fname) as fd:
            fd.write(lab_data)


def set_current_lab(lab_id):
    """
    creates a link to the cached lab to say it's current
    """

    cache_root = get_cache_root()
    fname = "{}/{}".format(cache_root, lab_id)
    if not os.path.exists(fname):
        raise FileNotFoundError("Failed to find cached lab for ID {}".format(lab_id))

    # This is supported on Windows and Unix as of Python 3.2
    # provided a new enough version of Windows.
    target = get_current_lab_link()
    clear_current_lab()
    os.symlink(fname, target)


def get_current_lab():
    """
    gets the current lab on which we're operating
    """

    lname = get_current_lab_link()
    if os.path.exists(lname):
        return os.path.basename(os.readlink(lname))

    return None


def clear_current_lab(lab_id=None):
    """
    unsets the current lab
    """
    lname = get_current_lab_link()
    if os.path.exists(lname):
        if lab_id is None or lab_id == get_current_lab():
            os.remove(lname)


def _extract_configuration_task(node):
    """
    concurrent task to extract a config from a node
    """
    if node.is_booted():
        node.extract_configuration()


def extract_configurations(lab):
    """
    extract each node's configuration to its day-0 config
    """
    # The client library prints "API Error" warnings when a node doesn't support extraction.  Quiet these.
    logger = logging.getLogger("virl2_client.models.authentication")
    level = logger.getEffectiveLevel()
    logger.setLevel(logging.CRITICAL)
    with concurrent.futures.ThreadPoolExecutor(max_workers=20) as executor:
        future_nodes = {executor.submit(_extract_configuration_task, node): node for node in lab.nodes()}
        for fn in concurrent.futures.as_completed(future_nodes):
            n = future_nodes[fn]
            try:
                fn.result()
            except HTTPError as he:
                if he.response.status_code != 400:
                    # Ignore 400 as that typically means the node doesn't support config extraction.
                    click.secho("WARNING: Failed to extract configuration from node {}: {}".format(n.label, he), fg="yellow")
            except Exception as e:
                click.secho("WARNING: Failed to extract configuration from node {}: {}".format(n.label, e), fg="yellow")

    logger.setLevel(level)


def get_node_mgmt_ip(node):
    """
    attempt to get the management (external) IP for a node
    """
    mgmtip = None
    for i in node.interfaces():
        if i.discovered_ipv4 and len(i.discovered_ipv4) > 0:
            mgmtip = i.discovered_ipv4[0]
        elif i.discovered_ipv6 and len(i.discovered_ipv6) > 0:
            for i6 in i.discovered_ipv6:
                if not ipaddress.ip_address(i6).is_link_local:
                    mgmtip = i6
                    break

        if mgmtip:
            break

    return mgmtip


def get_cml_client(server, ignore=False):
    """
    Helper function to get a consistent CML client library object
    """

    ssl_verify = not ignore
    # The virl2_client library warns about some stuff which can be confusing.
    # This eliminates the warnings, but the level is reset back to the current level
    # Before returning.
    logger = logging.getLogger("virl2_client.virl2_client")
    level = logger.getEffectiveLevel()
    logger.setLevel(logging.ERROR)

    if not ignore and "CML_VERIFY_CERT" in server.config:
        if server.config["CML_VERIFY_CERT"].lower() == "false":
            ssl_verify = False
        else:
            ssl_verify = server.config["CML_VERIFY_CERT"]

    # Remove VIRL2 envvars if they exist.  These would conflict with the virlutils config.
    os.environ.pop("VIRL2_USER", None)
    os.environ.pop("VIRL2_PASS", None)
    os.environ.pop("VIRL2_URL", None)

    client = ClientLibrary(server.host, server.user, server.passwd, raise_for_auth_failure=True, ssl_verify=ssl_verify)
    logger.setLevel(level)

    return client


def get_command():
    """
    Get the command that invoked virlutils.
    It is done this way to avoid being set to setup.py during test.
    """

    command = "virl"
    if os.path.basename(sys.argv[0]) == "cml":
        command = "cml"

    return command



import tabulate
import click


def sync_table(sync_result):
    click.secho("""
    NSO Sync Report
    """)
    headers = ["Device", "Result"]
    table = list()

    sync_results = sync_result['tailf-ncs:output']['sync-result']
    for item in sync_results:
        tr = list()
        tr.append(item['device'])

        result = item['result']
        if result is True:
            result = "SUCCESS"
            color = 'green'
        else:
            result = "FAILED"
            color = 'red'

        tr.append(click.style(result, fg=color))
        table.append(tr)
    # wrap the output in this try/except block as some terminals
    # may have problem with the 'fancy_grid'
    try:
        click.echo(tabulate.tabulate(table, headers, tablefmt="fancy_grid"))
    except UnicodeEncodeError:
        click.echo(tabulate.tabulate(table, headers, tablefmt="grid"))



import click
from virl.api import VIRLServer
from virl.helpers import get_cml_client


@click.command()
@click.option(
    "--confirm/--no-confirm", show_default=False, default=True, help="Do not prompt for confirmation (default: prompt)", required=False,
)
def deregister(confirm):
    """
    deregister the Smart License
    """
    server = VIRLServer()
    client = get_cml_client(server)
    licensing = client.licensing

    ret = "y"
    if confirm:
        ret = input("Are you sure you want to deregister [y/N]? ")
        if not ret.lower().startswith("y"):
            click.secho("Not deregistering")
            exit(0)

    try:
        licensing.deregister()
    except Exception as e:
        click.secho("Failed to deregister with Smart Licensing: {}".format(e), fg="red")
        exit(1)



import click
from virl.cli.wipe.node.commands import node as nodec
from virl.cli.wipe.lab.commands import lab as labc


@click.group()
def wipe():
    """
    wipe a lab or nodes within a lab
    """
    pass


wipe.add_command(labc, name="lab")
wipe.add_command(nodec, name="node")



from . import BaseCMLTest
from click.testing import CliRunner
import requests_mock
import os


class CMLTestDown(BaseCMLTest):
    def setup_mocks(self, m):
        super().setup_mocks(m)
        m.put(self.get_api_path("labs/{}/stop".format(self.get_test_id())), json="STOPPED")
        m.get(self.get_api_path("labs/{}/check_if_converged".format(self.get_test_id())), json=True)

    def test_cml_down(self):
        with requests_mock.Mocker() as m:
            # Mock the request to return what we expect from the API.
            self.setup_mocks(m)
            virl = self.get_virl()
            runner = CliRunner()
            result = runner.invoke(virl, ["down"])
            self.assertEqual(0, result.exit_code)

    def test_cml_down_by_name(self):

        try:
            os.remove(".virl/current_cml_lab")
        except OSError:
            pass

        with requests_mock.Mocker() as m:
            # Mock the request to return what we expect from the API.
            self.setup_mocks(m)
            virl = self.get_virl()
            runner = CliRunner()
            result = runner.invoke(virl, ["down", "--lab-name", self.get_test_title()])
            self.assertEqual(0, result.exit_code)

    def test_cml_down_by_id(self):

        try:
            os.remove(".virl/current_cml_lab")
        except OSError:
            pass

        with requests_mock.Mocker() as m:
            # Mock the request to return what we expect from the API.
            self.setup_mocks(m)
            virl = self.get_virl()
            runner = CliRunner()
            result = runner.invoke(virl, ["down", "--id", self.get_test_id()])
            self.assertEqual(0, result.exit_code)



import click
from virl.api import VIRLServer, ViewerPlugin, NoPluginError
from virl.cli.views import image_list_table
from virl.helpers import get_cml_client


@click.command()
@click.option("--image", default=None)
def ls(**kwargs):
    """
    list all images or the details of a specific image
    """

    image = kwargs.get("image")
    server = VIRLServer()
    client = get_cml_client(server)
    pl = None

    # Regardless of the argument, we have to get all the flavors
    # In the case of no arg, we print them all.
    # In the case of an arg, we have to go back and get details.
    defs = client.definitions.image_definitions()

    try:
        pl = ViewerPlugin(viewer="image_def")
    except NoPluginError:
        pass

    if image:
        for f in list(defs):
            if f["name"] == image:
                if pl:
                    pl.visualize(image_defs=[f])
                else:
                    image_list_table([f])
                break
    else:
        if pl:
            pl.visualize(image_defs=defs)
        else:
            image_list_table(defs)



from . import BaseCMLTest
from click.testing import CliRunner
import requests_mock


class CMLIdTest(BaseCMLTest):
    def test_cml_id(self):
        virl = self.get_virl()
        with requests_mock.Mocker() as m:
            # Mock the request to return what we expect from the API.
            self.setup_mocks(m)
            runner = CliRunner()
            result = runner.invoke(virl, ["id"])
            self.assertEqual("{} (ID: {})\n".format(self.get_test_title(), self.get_test_id()), result.output)



import requests
from .credentials import get_prop, _get_password, _get_from_user
from jinja2 import Environment


class NSO(object):
    __rfc8040 = False
    __nso_username = None
    __nso_password = None
    __nso_host = None

    def __init__(self):
        self.__get_credentials()
        self.__check_for_rfc8040()

    def __get_credentials(self):  # pragma: no cover

        configurable_props = ["NSO_HOST", "NSO_USERNAME", "NSO_PASSWORD"]
        config = dict()

        for p in configurable_props:
            val = get_prop(p)
            if val:
                config[p] = val

        if not config.get("NSO_HOST"):
            config["NSO_HOST"] = _get_from_user("Enter NSO IP/Hostname: ")

        if not config.get("NSO_USERNAME"):
            config["NSO_USERNAME"] = _get_from_user("Enter NSO username: ")

        if not config.get("NSO_PASSWORD"):
            config["NSO_PASSWORD"] = _get_password("Enter NSO password: ")

        self.__nso_host = config["NSO_HOST"]
        self.__nso_username = config["NSO_USERNAME"]
        self.__nso_password = config["NSO_PASSWORD"]

    def __check_for_rfc8040(self):
        url = "http://{}:8080/.well-known/host-meta".format(self.__nso_host)
        headers = {"Accept": "application/json"}
        rfc8040 = False

        try:
            response = requests.request("GET", url, auth=(self.__nso_username, self.__nso_password), headers=headers)
            response.raise_for_status()
        except Exception:
            pass
        else:
            json = response.json()
            if "links" in json and "restconf" in json["links"]:
                for u in json["links"]["restconf"]:
                    if u["href"] == "/restconf":
                        rfc8040 = True

        self.__rfc8040 = rfc8040

    def __build_ned_vars(self):
        url = "http://{}:8080".format(self.__nso_host)
        headers = dict()
        nurl = ""
        murl = ""

        if not self.__rfc8040:
            nurl = url + "/api/config/devices/ned-ids/ned-id"
            murl = url + "/api/config/modules-state/module"  # XXX: Is this correct for pre-RFC8040 NSO?
            headers = {"Content-Type": "application/vnd.yang.operation+json", "Accept": "application/vnd.yang.operation+json"}
        else:
            nurl = url + "/restconf/data/tailf-ncs:devices/ned-ids/ned-id"
            murl = url + "/restconf/data/ietf-yang-library:modules-state/module"
            headers = {"Content-Type": "application/yang-data+json", "Accept": "application/yang-data+json"}

        response = requests.request("GET", nurl, auth=(self.__nso_username, self.__nso_password), headers=headers)
        neds = response.json()
        ned_patterns = {
            "cisco-ios-": "IOS",
            "cisco-iosxr-": "XR",
            "cisco-nx-": "NX",
            "cisco-asa-": "ASA",
        }
        # Set some "sane" defaults.  These are at least the same as what the older
        # virlutils had.
        ned_vars = {
            "NX_PREFIX": "cisco-nx-id",
            "NX_NED_ID": "cisco-nx",
            "NX_NAMESPACE": "http://tail-f.com/ned/cisco-nx-id",
            "XR_PREFIX": "cisco-ios-xr-id",
            "XR_NED_ID": "cisco-ios-xr",
            "XR_NAMESPACE": "http://tail-f.com/ned/cisco-ios-xr-id",
            "IOS_PREFIX": "ios-id",
            "IOS_NED_ID": "cisco-ios",
            "IOS_NAMESPACE": "urn:ios-id",
            "ASA_PREFIX": "asa-id",
            "ASA_NED_ID": "cisco-asa",
            "ASA_NAMESPACE": "http://cisco.com/ned/asa-id",
        }

        response = requests.request("GET", murl, auth=(self.__nso_username, self.__nso_password), headers=headers)
        modules = response.json()

        for ned in neds["tailf-ncs:ned-id"]:
            (prefix, nid) = ned["id"].split(":")
            for pattern, platform in ned_patterns.items():
                # FIXME: This only uses the first encountered matching NED ID.
                # If one has multiple NEDs, the desirned NED may not be ideal.
                if nid.startswith(pattern):
                    ned_vars["{}_PREFIX".format(platform)] = prefix
                    ned_vars["{}_NED_ID".format(platform)] = nid
                    for module in modules["ietf-yang-library:module"]:
                        if module["name"] == nid:
                            ned_vars["{}_NAMESPACE".format(platform)] = module["namespace"]
                            break
                    break

        return ned_vars

    def perform_sync_from(self):
        url = "http://{}:8080".format(self.__nso_host)
        headers = dict()

        if not self.__rfc8040:
            url += "/api/running/devices/_operations/sync-from"
            headers = {"Content-Type": "application/vnd.yang.operation+json", "Accept": "application/vnd.yang.operation+json"}
        else:
            url += "/restconf/operations/tailf-ncs:devices/tailf-ncs:sync-from"
            headers = {"Content-Type": "application/yang-data+json", "Accept": "application/yang-data+json"}

        response = requests.request("POST", url, auth=(self.__nso_username, self.__nso_password), headers=headers)
        return response

    # def perform_sync_to():
    #     nso_host, nso_username, nso_password = get_credentials()
    #     url = "http://{}:8080".format(nso_host)
    #     url = url + "/api/running/devices/_operations/sync-to".format(nso_host)
    #     headers = {'Content-Type': "application/vnd.yang.operation+json",
    #                'Accept': "application/vnd.yang.operation+json"}
    #
    #     response = requests.request("POST",
    #                                 url,
    #                                 auth=(nso_username, nso_password),
    #                                 headers=headers)
    #     return response

    def update_devices(self, xml_payload):
        url = "http://{}:8080".format(self.__nso_host)
        headers = dict()

        if not self.__rfc8040:
            url += "/api/config/devices/"
            headers = {"Content-Type": "application/vnd.yang.data+xml"}
        else:
            url += "/restconf/data/tailf-ncs:devices"
            headers = {"Content-Type": "application/yang-data+xml"}

        ned_vars = self.__build_ned_vars()
        env = Environment()
        template = env.from_string(source=xml_payload, globals=ned_vars)

        response = requests.request("PATCH", url, auth=(self.__nso_username, self.__nso_password), data=template.render(), headers=headers)

        return response



class MockGitHub(object):
    @staticmethod
    def get_topology(req, ctx):
        with open("tests/v2/static/fake_repo_topology.yaml", "r") as fh:
            response = fh.read()
        return response



import click
import os
from virl.api import VIRLServer
from virl.helpers import get_cml_client


@click.command()
@click.argument("image", nargs=1)
@click.option("-f", "--filename", required=True, metavar="<filename>", help="path to the local image definition file")
def definition(image, filename):
    """
    import an image definition
    """

    server = VIRLServer()
    client = get_cml_client(server)

    if not os.path.isfile(filename):
        click.secho("Image definition file {} does not exist or is not a file", fg="red")
        exit(1)
    else:
        defs = client.definitions
        contents = None

        with open(filename, "r") as fd:
            contents = fd.read()

        try:
            defs.upload_image_definition(image, contents)
        except Exception as e:
            click.secho("Failed to import image definition for {}: {}".format(image, e), fg="red")
            exit(1)



import click
from virl.api import VIRLServer
from subprocess import call
from virl import helpers
from virl2_client.exceptions import NodeNotFound
from virl.helpers import get_cml_client, safe_join_existing_lab, get_current_lab, get_command


@click.command()
@click.argument("node", nargs=1)
def stop(node):
    """
    stop a node
    """
    server = VIRLServer()
    client = get_cml_client(server)

    current_lab = get_current_lab()
    if current_lab:
        lab = safe_join_existing_lab(current_lab, client)
        if lab:
            try:
                node_obj = lab.get_node_by_label(node)

                if node_obj.is_active():
                    node_obj.stop(wait=True)
                    click.secho("Stopped node {}".format(node_obj.label))
                else:
                    click.secho("Node {} is already stopped".format(node_obj.label), fg="yellow")
            except NodeNotFound:
                click.secho("Node {} was not found in lab {}".format(node, current_lab), fg="red")
                exit(1)
        else:
            click.secho("Unable to find lab {}".format(current_lab), fg="red")
            exit(1)
    else:
        click.secho("No current lab set", fg="red")
        exit(1)


@click.command()
@click.argument("node", nargs=-1)
def stop1(node):
    """
    stop a node
    """
    if len(node) == 2:
        # we received env and node name
        env = node[0]
        running = helpers.check_sim_running(env)
        node = node[1]
    elif len(node) == 1:
        # assume default env
        env = "default"
        running = helpers.check_sim_running(env)
        node = node[0]
    else:
        exit(call([get_command(), "stop", "--help"]))

    if running:
        sim_name = running
        server = VIRLServer()
        resp = server.stop_node(sim_name, node)
        if resp.ok:
            click.secho("Stopped node {}".format(node))
        else:
            click.secho("Error Stopping Node {}: {}".format(node, resp), fg="red")



from . import BaseTest
from virl.api import VIRLServer


class Tests(BaseTest):
    def test_create_virl_server(self):
        server = VIRLServer()
        self.assertIsInstance(server, VIRLServer)

    def test_user(self):
        server = VIRLServer()
        server.user = "foo"
        self.assertEqual(server.user, "foo")

    def test_host(self):
        server = VIRLServer()
        server.host = "notreal"
        self.assertEqual(server.host, "notreal")

    def test_passwd(self):
        server = VIRLServer()
        server.passwd = "notreal"
        self.assertEqual(server.passwd, "notreal")



from . import BaseTest
from click.testing import CliRunner
import requests_mock
from virl.cli.main import virl
import os


class Tests(BaseTest):

    def test_virl_down(self):
        with requests_mock.mock() as m:
            # Mock the request to return what we expect from the API.
            down_url = 'http://localhost:19399/simengine/rest/stop/TEST_ENV'
            m.get(down_url, json=self.mock_down_response())
            runner = CliRunner()
            result = runner.invoke(virl, ["down"])
            self.assertEqual(0, result.exit_code)

    def test_virl_down_by_name(self):

        try:
            os.remove('.virl/default/id')
        except OSError:
            pass

        with requests_mock.mock() as m:
            # Mock the request to return what we expect from the API.
            down_url = 'http://localhost:19399/simengine/rest/stop/FOO'
            m.get(down_url, json=self.mock_down_response())
            runner = CliRunner()
            result = runner.invoke(virl, ["down", "--sim-name", "FOO"])
            self.assertEqual(0, result.exit_code)

    def mock_down_response(self):
        response = 'SUCCESS'
        return response



from .api import VIRLServer  # noqa
from .cml import CachedLab  # noqa
from .plugin import load_plugins, check_valid_plugin, Plugin, ViewerPlugin, CommandPlugin, GeneratorPlugin, NoPluginError  # noqa



import requests_mock
from . import BaseTest
from click.testing import CliRunner
from virl.cli.main import virl


class SearchTests(BaseTest):

    @requests_mock.mock()
    def test_virl_search(self, m):
        m.get('https://api.github.com/orgs/virlfiles/repos',
              json=self.mock_response())
        runner = CliRunner()
        result = runner.invoke(virl, ["search"])
        self.assertEqual(0, result.exit_code)

    @requests_mock.mock()
    def test_virl_search_with_query_name(self, m):
        m.get('https://api.github.com/orgs/virlfiles/repos',
              json=self.mock_response())
        runner = CliRunner()
        result = runner.invoke(virl, ["search", "ios"])
        self.assertEqual(0, result.exit_code)

    @requests_mock.mock()
    def test_virl_search_with_query_descr(self, m):
        # Mock the request to return what we expect from the API.
        m.get('https://api.github.com/orgs/virlfiles/repos',
              json=self.mock_response())
        runner = CliRunner()
        result = runner.invoke(virl, ["search", "hello"])
        self.assertEqual(0, result.exit_code)

    def mock_response(self):
        response = [{
                    "name": "2-ios-router",
                    "full_name": "virlfiles/2-ios-router",
                    "description": "hello world virlfile",
                    "stargazers_count": 344
                    }]
        return response



import click
from virl.cli.license.renew.registration.commands import registration as registrationc
from virl.cli.license.renew.authorization.commands import authorization as authorizationc


@click.group()
def renew():
    """
    renew registration or authorization
    """
    pass


renew.add_command(registrationc, name="registration")
renew.add_command(authorizationc, name="authorization")



import click
from virl.api import VIRLServer
import subprocess


@click.command()
def cockpit():
    """
    opens the Cockpit UI
    """
    server = VIRLServer()
    url = "https://{}:9090".format(server.host)
    subprocess.Popen(["open", url])



import click
from virl.api import VIRLServer
from requests.exceptions import HTTPError


@click.command()
@click.argument('flavor')
@click.option('--memory', default=None)
@click.option('--vcpus', default=None)
@click.option('--name', default=None)
def update(**kwargs):
    """
    update a specific flavor
    """

    flavor = kwargs.get('flavor')
    memory = kwargs.get('memory')
    vcpus = kwargs.get('vcpus')
    name = kwargs.get('name')
    server = VIRLServer(port=80)

    # Does the flavor exist?
    try:
        id = server.get_flavor_id(flavor)
        r = server.get_flavors(id)
    except HTTPError as err:
        print("\nFailed to locate flavor '{}':".format(flavor))
        print("\n\t{}".format(err.response.reason))
        return

    f = r

    if not memory:
        memory = f['ram']

    if not vcpus:
        vcpus = f['vcpus']

    if not name:
        name = f['name']

    r = server.delete_flavor(flavor)
    r = server.add_flavor(flavor=name,
                          memory=memory,
                          vcpus=vcpus)

    print("Original Flavor")
    print("\t   ID: {}".format(f['id']))
    print("\t Name: {}".format(f['name']))
    print("\t  RAM: {}".format(f['ram']))
    print("\tVCPUS: {}".format(f['vcpus']))

    print("New Flavor")
    print("\t   ID: {}".format(r['id']))
    print("\t Name: {}".format(r['name']))
    print("\t  RAM: {}".format(r['ram']))
    print("\tVCPUS: {}".format(r['vcpus']))



import click
from virl.api import VIRLServer
from virl.cli.views import sync_table
from virl import helpers
from virl.helpers import get_cml_client, safe_join_existing_lab, get_current_lab
from virl.generators import nso_payload_generator1, nso_payload_generator
from virl.api.nso import NSO


@click.command()
@click.option("--output", "-o", help="just dump the payload to file without sending")
@click.option("--syncfrom/--no-syncfrom", default=False, help="Perform sync-from after updating devices")
def nso(syncfrom, **kwargs):
    """
    generate nso inventory
    """

    server = VIRLServer()
    client = get_cml_client(server)

    current_lab = get_current_lab()
    if current_lab:
        lab = safe_join_existing_lab(current_lab, client)
        if lab:

            if kwargs.get("output"):
                file_name = kwargs.get("output")
            else:
                file_name = None

            inv = nso_payload_generator(lab, server)

            if inv:
                if file_name:
                    click.secho("Writing {}".format(file_name))
                    with open(file_name, "w") as fd:
                        fd.write(inv)
                else:
                    click.secho("Updating NSO....")
                    nso_obj = NSO()
                    nso_response = nso_obj.update_devices(inv)
                    if nso_response.ok:
                        click.secho("Successfully added CML devices to NSO")
                    else:
                        click.secho("Error updating NSO: ", fg="red")
                        click.secho(nso_response.text)
                    if syncfrom:
                        resp = nso_obj.perform_sync_from()
                        sync_table(resp.json())
            else:
                click.secho("Failed to get inventory data", fg="red")
                exit(1)
        else:
            click.secho("Failed to find running lab {}".format(current_lab), fg="red")
            exit(1)
    else:
        click.secho("Current lab is not set", fg="red")
        exit(1)


@click.command()
@click.argument("env", default="default")
@click.option("--output", "-o", help="just dump the payload to file without sending")
@click.option("--syncfrom/--no-syncfrom", default=False, help="Perform sync-from after updating devices")
# @click.option('--syncto/--no-syncto', default=False,
#               help="Perform sync-to afgter updating devices")
def nso1(env, syncfrom, **kwargs):
    """
    generate nso inventory
    """

    if kwargs.get("output"):
        # user specified output filename
        file_name = kwargs.get("output")
    else:
        # writes to <env>.json by default
        file_name = None

    running = helpers.check_sim_running(env)
    if running:
        sim_name = running
        server = VIRLServer()
        roster = server.get_sim_roster(sim_name)
        # sim_name = "topology-fpyHFs"
        virl_data = server.export(sim_name, ip=True).content
        interfaces = server.get_interfaces(sim_name).json()

        payload = nso_payload_generator1(sim_name, virl_data, roster=roster, interfaces=interfaces)

        if file_name:  # pragma: no cover
            click.secho("Writing payload to {}".format(file_name))
            with open(file_name, "w") as payload_file:

                payload_file.write(payload)
        else:
            click.secho("Updating NSO....")
            nso_obj = NSO()
            nso_response = nso_obj.update_devices(payload)
            if nso_response.ok:
                click.secho("Successfully added VIRL devices to NSO")
            else:
                click.secho("Error updating NSO: ", fg="red")
                click.secho(nso_response.text)
            if syncfrom:
                resp = nso_obj.perform_sync_from()
                sync_table(resp.json())

    else:
        click.secho("couldnt generate testbed for for env: {}".format(env), fg="red")



import click
from virl.api import VIRLServer
from virl import helpers
import subprocess


@click.command()
@click.argument('env', default='default')
def uwm1(env):
    """
    opens UWM for the sim
    """
    server = VIRLServer()
    running = helpers.check_sim_running(env)

    if running:
        sim_name = running
        # luanches uwm
        url = "http://{}:{}@{}/simulation/{}/{}"
        url = url.format(server.user, server.passwd,
                         server.host, server.user, sim_name)
        subprocess.Popen(['open', url])



import click
from virl.api import VIRLServer
from virl import __version__
from virl.helpers import get_cml_client


@click.command()
def version():
    """
    version information
    """
    server = VIRLServer()
    client = get_cml_client(server)
    server_version = "Unknown"
    try:
        response = client.session.get(client._base_url + "system_information")
        response.raise_for_status()
        server_version = response.json()["version"]
    except Exception:
        pass
    virlutils_version = __version__
    click.secho("virlutils Version: {}".format(virlutils_version))
    click.secho("CML Controller Version: {}".format(server_version))


@click.command()
def version1():
    """
    version information
    """
    server = VIRLServer()
    virlutils_version = __version__
    server_version = server.get_version().get("virl-version")
    click.secho("virlutils Version: {}".format(virlutils_version))
    click.echo("VIRL Core Version: {}".format(server_version))



import click
from virl.api import VIRLServer
from virl.helpers import get_cml_client


@click.command()
def registration():
    """
    renew Smart License registration
    """
    server = VIRLServer()
    client = get_cml_client(server)
    licensing = client.licensing

    try:
        licensing.register_renew()
    except Exception as e:
        click.secho("Failed to renew registration: {}".format(e), fg="red")
        exit(1)



import click
from virl.api import VIRLServer, load_plugins, CommandPlugin, Plugin, check_valid_plugin, NoPluginError
from virl.helpers import get_default_plugin_dir
import requests
from requests.packages.urllib3.exceptions import InsecureRequestWarning
import warnings
import virl2_client
import traceback
from virl.helpers import get_cml_client, get_command
from .console.commands import console, console1  # noqa: F401
from .nodes.commands import nodes, nodes1  # noqa: F401
from .logs.commands import logs1  # noqa: F401
from .up.commands import up, up1  # noqa: F401
from .use.commands import use, use1  # noqa: F401
from .down.commands import down, down1  # noqa: F401
from .ls.commands import ls, ls1  # noqa: F401
from .save.commands import save, save1  # noqa: F401
from .telnet.commands import telnet, telnet1  # noqa: F401
from .ssh.commands import ssh, ssh1  # noqa: F401
from .generate import generate, generate1, init_generators  # noqa: F401
from .start.commands import start, start1  # noqa: F401
from .stop.commands import stop, stop1  # noqa: F401
from .pull.commands import pull, pull1  # noqa: F401
from .search.commands import search  # noqa: F401
from .swagger.commands import swagger1  # noqa: F401
from .uwm.commands import uwm1  # noqa: F401
from .viz.commands import viz1  # noqa: F401
from .id.commands import lid, sid  # noqa: F401
from .version.commands import version, version1  # noqa: F401
from .flavors import flavors1  # noqa: F401
from .definitions import definitions  # noqa: F401
from .cockpit.commands import cockpit  # noqa: F401
from .wipe import wipe  # noqa: F401
from .extract.commands import extract  # noqa: F401
from .clear.commands import clear  # noqa: F401
from .ui.commands import ui  # noqa: F401
from .license import license  # noqa: F401
from .rm.commands import rm  # noqa: F401
from .command.commands import command  # noqa: F401


class CatchAllExceptions(click.Group):
    def __call__(self, *args, **kwargs):
        try:
            return self.main(*args, **kwargs)
        except Exception as exc:
            click.secho("Exception raised while running your command", fg="red")
            if not virl.debug:
                click.secho(
                    "Please re-run as '{} --debug ...' and collect the output before opening an issue".format(get_command()),
                    fg="red",
                )
            else:
                click.secho("Please open an issue and provide this output:", fg="red")
            click.secho("%s" % exc, fg="red")
            if virl.debug:
                click.secho(traceback.format_exc(), fg="red")
            exit(1)


@click.group(cls=CatchAllExceptions)
@click.option(
    "--debug/--no-debug",
    default=False,
    help="Print any debugging output.",
    required=False,
)
def virl(**kwargs):
    if kwargs.get("debug"):
        virl.debug = True

    # We need to pull this out or subcommands fail.
    kwargs.pop("debug", None)


# Shall we print any debugging output?
virl.debug = False


def __get_server_ver():
    """
    Taste a VIRL/CML server and try and determine its version.
    This tries the 2.x flow and assumes 1.x if that flow fails in an
    unexpected way.  The reason for this is that compatibility with 1.x
    is stressed, but the code has been factored in a way so that the 1.x
    code can be removed when the time is right.

    Returns:
        string: Either '1' for VIRL/CML 1.x or the empty string for CML 2+
    """
    res = ""
    try:
        server = VIRLServer()
        if "CML2_PLUS" not in server.config:
            # If the user hasn't explicitly said they are on the CML 2+, then
            # attempt to guess the server version.

            # We don't care about cert validation here.  If this is a CML server,
            # we'll fail validation later anyway.
            #
            # Because of that, pass obviously bogus credentials.  The login will fail
            # in a predictable way if this is a CML server.
            requests.packages.urllib3.disable_warnings(InsecureRequestWarning)
            r = requests.get("https://{}/".format(server.host), verify=False)
            warnings.simplefilter("default", InsecureRequestWarning)
            r.raise_for_status()
            # While one could have a user called virutils-test, the user must have a password.
            # So if we send an empty password, that will fail with a known error.
            server.user = "virlutils-test"
            server.passwd = ""
            get_cml_client(server, ignore=True)
    except virl2_client.InitializationError:
        # The client library will raise this error if it encounters an authorization failure.
        pass
    except Exception:
        # Any other error likely means a VIRL/CML 1.x host.
        res = "1"

    return res


def __init_plugins():
    """
    Scan a set of plugin directories and load them if any are found.
    Plugins come in one of three types: command, generator, viewer.
    In general, plugins can override base functionality.
    """

    server = VIRLServer()
    plugin_dirs = server.config.get("CML_PLUGIN_PATH", "")
    if plugin_dirs != "":
        plugin_dirs += ":"

    plugin_dirs += get_default_plugin_dir()
    load_plugins(plugin_dirs)

    for cmd in Plugin.get_plugins("command"):
        try:
            pl = CommandPlugin(command=cmd)
        except NoPluginError:
            continue
        if not check_valid_plugin(pl, pl.run, "run"):
            click.secho(
                "ERROR: Malformed plugin for command {}.  The `run` method must be static and a click.command".format(cmd), fg="red"
            )
            Plugin.remove_plugin("command", cmd)
        else:
            virl.add_command(pl.run, name=cmd)

    # initialize the 'generate' command arguments after we've loaded
    # any plugins.  Else the plugins will not be available
    init_generators()


__server_ver = __get_server_ver()

if __server_ver == "1":
    virl.add_command(uwm1, name="uwm")
    virl.add_command(flavors1, name="flavors")
    virl.add_command(logs1, name="logs")
    virl.add_command(swagger1, name="swagger")
    virl.add_command(viz1, name="viz")
    virl.add_command(sid, name="id")
else:
    virl.add_command(cockpit)
    virl.add_command(definitions)
    virl.add_command(wipe)
    virl.add_command(extract)
    virl.add_command(clear)
    virl.add_command(ui)
    virl.add_command(license)
    virl.add_command(rm)
    virl.add_command(lid, name="id")
    virl.add_command(command)

virl.add_command(search)

__sub_commands = [
    "console",
    "nodes",
    "up",
    "down",
    "ls",
    "use",
    "save",
    "telnet",
    "ssh",
    "generate",
    "start",
    "stop",
    "pull",
    "version",
]

for cmd in __sub_commands:
    virl.add_command(globals()[cmd + __server_ver], name=cmd)

# Load plugins, but only for CML 2+
if __server_ver != "1":
    __init_plugins()

if __name__ == "__main__":
    virl()  # pragma: no cover



import tabulate
import click


def log_table(log_entries):
    click.secho("Here is a list of recent log entries")
    # sort by date
    log_entries = sorted(log_entries, key=lambda k: k['id'])
    headers = ["Timestamp", "Severity", "Message"]
    table = list()

    for log in log_entries:
        tr = list()
        tr.append(log['time'])
        level = log['level']
        if level == 'INFO':
            color = "green"
        else:
            color = 'yellow'
        tr.append(click.style(level, fg=color))
        tr.append(log['message'])
        table.append(tr)
    # wrap the output in this try/except block as some terminals
    # may have problem with the 'fancy_grid'
    try:
        click.echo(tabulate.tabulate(table, headers, tablefmt="fancy_grid"))
    except UnicodeEncodeError:
        click.echo(tabulate.tabulate(table, headers, tablefmt="grid"))



import click
from virl.cli.definitions.nodes import nodes
from virl.cli.definitions.images import images


@click.group()
def definitions():
    """
    manage image and node definitions
    """
    pass


definitions.add_command(nodes)
definitions.add_command(images)



import click
from virl.api import VIRLServer
from virl.helpers import get_cml_client, get_current_lab, safe_join_existing_lab
import subprocess


@click.command()
def ui():
    """
    opens the Workbench for the current lab
    """
    server = VIRLServer()
    client = get_cml_client(server)

    current_lab = get_current_lab()
    if current_lab:
        lab = safe_join_existing_lab(current_lab, client)
        if lab:
            url = "https://{}/lab/{}".format(server.host, current_lab)
            subprocess.Popen(["open", url])



import click
from virl.api import VIRLServer, ViewerPlugin, NoPluginError
from virl.cli.views import license_features_table
from virl.helpers import get_cml_client


@click.command()
def show():
    """
    display license details
    """
    server = VIRLServer()
    client = get_cml_client(server)
    licensing = client.licensing

    try:
        license = licensing.features()
    except Exception as e:
        click.secho("Failed to get license features: {}".format(e), fg="red")
        exit(1)
    else:
        try:
            pl = ViewerPlugin(viewer="license_feature")
            pl.visualize(features=license)
        except NoPluginError:
            license_features_table(license)



import os
from . import BaseTest
from click.testing import CliRunner
from virl.cli.main import virl
from virl.swagger.app import app

try:
    from unittest.mock import patch
except ImportError:
    from mock import patch


class Tests(BaseTest):

    @patch("virl.cli.swagger.commands.subprocess.Popen", auto_spec=False)
    def test_virl_swagger_start(self, call_mock):
        runner = CliRunner()
        runner.invoke(virl, ["swagger", "start"])
        url = 'http://localhost:5000'
        call_mock.assert_called_with(['open', url])

    def test_virl_swagger_status(self):
        with open('/tmp/virl_swagger.port', 'w') as fh:
            fh.write('1234')
        with open('/tmp/virl_swagger.pid', 'w') as fh:
            fh.write('1234')
        runner = CliRunner()
        result = runner.invoke(virl, ["swagger", "status"])
        expected_output = 'VIRL swagger UI is running at http://localhost:1234'
        self.assertIn(expected_output, result.output)

    @patch("virl.cli.swagger.commands.subprocess.Popen", auto_spec=False)
    def test_virl_swagger_stop(self, call_mock):
        runner = CliRunner()
        result = runner.invoke(virl, ["swagger", "stop"])
        print(result.output)
        call_mock.assert_called_with('kill $(cat /tmp/virl_swagger.pid)',
                                     shell=True)


class SwaggerFlaskApp(BaseTest):
    def setUp(self):
        os.environ['VIRL_SWAGGER_HOST'] = 'localhost'
        os.environ['VIRL_SWAGGER_USERNAME'] = 'guest'
        os.environ['VIRL_SWAGGER_PASSWORD'] = 'guest'
        os.environ['VIRL_SWAGGER_PORT'] = '1234'
        self.app = app
        app.testing = True
        self.app = app.test_client()

    def test_swagger_ui(self):
        resp = self.app.get('/')
        expected_data = b'ui.preauthorizeBasic("basicAuth", "guest", "guest")'
        self.assertEqual(resp.status_code, 200)
        self.assertIn(expected_data, resp.data)

    def test_swagger_spec(self):
        resp = self.app.get('/swagger.json')
        expected_data = b'"host": "localhost:1234",'
        self.assertEqual(resp.status_code, 200)
        self.assertIn(expected_data, resp.data)

    def tearDown(self):
        pass
        # os.close(self.db_fd)
        # os.unlink(flaskr.app.config['DATABASE'])



import os
import unittest
from click.testing import CliRunner
import functools
import sys
import traceback
import pdb


def setup_environ():
    os.environ["VIRL_HOST"] = "localhost"
    os.environ["VIRL_USERNAME"] = "guest"
    os.environ["VIRL_PASSWORD"] = "guest"
    os.environ["NSO_HOST"] = "localhost"
    os.environ["NSO_USERNAME"] = "admin"
    os.environ["NSO_PASSWORD"] = "admin"


setup_environ()
# This is done specifically to avoid prompting during import
# while running "make test" or "make coverage".
from virl.cli.main import virl  # noqa: E402


def debug_on(*exceptions):
    if not exceptions:
        exceptions = (AssertionError,)

    def decorator(f):
        @functools.wraps(f)
        def wrapper(*args, **kwargs):
            try:
                return f(*args, **kwargs)
            except exceptions:
                info = sys.exc_info()
                traceback.print_exception(*info)
                pdb.post_mortem(info[2])

        return wrapper

    return decorator


class BaseTest(unittest.TestCase):
    def setUp(self):
        # Only doing this because we don't have a better way of controlling
        # injection of VIRL_HOST
        setup_environ()
        runner = CliRunner()
        runner.invoke(virl, ["use", "TEST_ENV"])



import click
from virl.api import VIRLServer
import time
from virl.helpers import get_cml_client, safe_join_existing_lab, get_current_lab
from virl2_client import NodeNotFound


@click.command()
@click.option("--force/--no-force", "-f", default=False, required=False, help="Stop a node (if it's started) then wipe it (default: False)")
@click.option(
    "--confirm/--no-confirm",
    show_default=False,
    default=True,
    help="Do not prompt for confirmation (default: prompt)",
    required=False,
)
@click.argument("node", nargs=1)
def node(node, force, confirm):
    """
    wipe a node
    """
    server = VIRLServer()
    client = get_cml_client(server)

    current_lab = get_current_lab()
    if current_lab:
        lab = safe_join_existing_lab(current_lab, client)
        if lab:
            try:
                node_obj = lab.get_node_by_label(node)
            except NodeNotFound:
                click.secho("Node {} was not found in lab {}".format(node, current_lab), fg="red")
                exit(1)

            if node_obj.is_active() and force:
                node_obj.stop()
                while node_obj.is_active():
                    time.sleep(1)

            if not node_obj.is_active():
                ret = "y"
                if confirm:
                    ret = input("Are you sure you want to wipe node {} [y/N]? ".format(node_obj.label))
                if ret.lower().startswith("y"):
                    node_obj.wipe(wait=True)
                    click.secho("Node {} wiped".format(node_obj.label))
                else:
                    click.secho("Not wiping node {}".format(node_obj.label))
            else:
                click.secho("Node {} is active; either stop it or use --force".format(node_obj.label), fg="red")
                exit(1)
        else:
            click.secho("Unable to find lab {}".format(current_lab), fg="red")
            exit(1)
    else:
        click.secho("No current lab set", fg="red")
        exit(1)



from . import BaseTest
from click.testing import CliRunner
import requests_mock
from virl.cli.main import virl
try:
    from unittest.mock import patch
except ImportError:
    from mock import patch


class Tests(BaseTest):

    def test_virl_start(self):
        with requests_mock.mock() as m:
            # Mock the request to return what we expect from the API.
            down_url = 'http://localhost:19399/simengine/rest/start/TEST_ENV'
            m.get(down_url, json=self.mock_down_response())
            runner = CliRunner()
            runner.invoke(virl, ["start", "router1"])

    def test_virl_start_env(self):
        with requests_mock.mock() as m:
            # Mock the request to return what we expect from the API.
            down_url = 'http://localhost:19399/simengine/rest/start/foo_env'
            m.get(down_url, json=self.mock_down_response())
            runner = CliRunner()
            runner.invoke(virl, ["start", "foo_env", "router1"])

    @patch("virl.cli.start.commands.call", auto_spec=False)
    def test_virl_start_no_node_given(self, call_mock):
        with requests_mock.mock() as m:
            # Mock the request to return what we expect from the API.
            down_url = 'http://localhost:19399/simengine/rest/stop/TEST_ENV'
            m.get(down_url, json=self.mock_down_response())
            runner = CliRunner()
            runner.invoke(virl, ["start"])
            call_mock.assert_called_with(['virl', 'start', '--help'])

    def mock_down_response(self):
        response = 'SUCCESS'
        return response



from jinja2 import Environment, PackageLoader
from virl.helpers import get_node_mgmt_ip


def sim_info(virl_xml, roster=None, interfaces=None, protocol="telnet"):
    """
    common inventory info accross yaml/ini
    """
    inventory = list()

    for node_long_name, device in roster.items():
        # make sure we have at least enough info to be useful
        if ("managementIP") in device and ("NodeName" in device):
            entry = dict()
            entry["node_long_name"] = node_long_name
            entry["address"] = device["managementIP"]
            entry["protocol"] = protocol
        # otherwise, skip this device
        else:
            continue

        name = device["NodeName"]
        entry["name"] = name
        entry["ned"] = "unknown"
        entry["ns"] = "unkown"
        # map node types to known NED's
        try:
            type = device["NodeSubtype"]
            if "NX" in type:
                entry["prefix"] = "{{ NX_PREFIX }}"
                entry["ned"] = "{{ NX_NED_ID }}"
                entry["ns"] = "{{ NX_NAMESPACE }}"
            elif "XR" in type:
                entry["prefix"] = "{{ XR_PREFIX }}"
                entry["ned"] = "{{ XR_NED_ID }}"
                entry["ns"] = "{{ XR_NAMESPACE }}"
            elif "CSR" in type or "IOS" in type:
                entry["prefix"] = "{{ IOS_PREFIX }}"
                entry["ned"] = "{{ IOS_NED_ID }}"
                entry["ns"] = "{{ IOS_NAMESPACE }}"
            elif "ASA" in type:
                entry["prefix"] = "{{ ASA_PREFIX }}"
                entry["ned"] = "{{ ASA_NED_ID }}"
                entry["ns"] = "{{ ASA_NAMESPACE }}"
                entry["protocol"] = "ssh"
        except KeyError:
            pass

        if entry.get("ned", None) not in ["unknown", None]:
            inventory.append(entry)

    return inventory


def lab_info(lab, server, protocol):
    """
    common inventory info across xml/json
    """
    inventory = list()

    for node in lab.nodes():
        mgmtip = get_node_mgmt_ip(node)

        if not mgmtip:
            continue

        name = node.label
        entry = dict()
        entry["address"] = mgmtip
        entry["protocol"] = protocol
        name = node.label
        entry["name"] = name
        entry["ned"] = "unknown"
        entry["ns"] = "unkown"

        # determine device/os type
        try:
            type = node.node_definition.lower()
            if "nx" in type:
                entry["prefix"] = "{{ NX_PREFIX }}"
                entry["ned"] = "{{ NX_NED_ID }}"
                entry["ns"] = "{{ NX_NAMESPACE }}"
            elif "xr" in type:
                entry["prefix"] = "{{ XR_PREFIX }}"
                entry["ned"] = "{{ XR_NED_ID }}"
                entry["ns"] = "{{ XR_NAMESPACE }}"
            elif "csr" in type or "ios" in type:
                entry["prefix"] = "{{ IOS_PREFIX }}"
                entry["ned"] = "{{ IOS_NED_ID }}"
                entry["ns"] = "{{ IOS_NAMESPACE }}"
            elif "asa" in type:
                entry["prefix"] = "{{ ASA_PREFIX }}"
                entry["ned"] = "{{ ASA_NED_ID }}"
                entry["ns"] = "{{ ASA_NAMESPACE }}"
        except KeyError:
            pass

        if entry.get("ned", None) not in ["unknown", None]:
            inventory.append(entry)

    return inventory


def render_xml_payload1(virl_xml, roster=None, interfaces=None, protocol="telnet"):
    """
    we need to merge information from multiple sources to generate all
    the required parameters for the inventory

    """
    env = Environment(loader=PackageLoader("virl"), trim_blocks=False)

    inventory = sim_info(virl_xml, roster=roster, interfaces=interfaces, protocol=protocol)
    # pass all available data to template for rendering, this can probably be
    # pruned back at some point
    xml = env.get_template("nso/xml_payload.j2").render(inventory=inventory)
    return xml


def render_payload(lab, server, protocol, style):
    env = Environment(loader=PackageLoader("virl"), trim_blocks=False)

    if style == "json":  # pragma: no cover
        raise NotImplementedError

    inventory = lab_info(lab, server, protocol)
    payload = env.get_template("nso/xml_payload.j2").render(inventory=inventory)
    return payload


def nso_payload_generator1(env, virl_data, roster, interfaces, style="xml", protocol="telnet"):
    """
    given a sim roster produces a payload file suitable for
    use with network services orchestrator

    """
    if style == "xml":
        xml_payload = render_xml_payload1(virl_data, roster, interfaces)
        return xml_payload
    elif style == "json":  # pragma: no cover
        raise NotImplementedError


def nso_payload_generator(lab, server, style="xml", protocol="ssh"):
    """
    given a lab produces a payload file suitable for
    use with network services orchestrator
    """
    payload = render_payload(lab, server, protocol, style)

    return payload




class MockVIRLServer:

    @classmethod
    def launch_simulation(cls):
        response = u'TEST_ENV'
        return response

    @classmethod
    def get_node_console(cls):
        sim_response = {
            u'router2': u'10.94.241.194:17002',
            u'router1': u'10.94.241.194:17001'
        }
        return sim_response

    @classmethod
    def get_sim_roster(cls):
        response = {
            u'guest|TEST_ENV|virl|router1': {
                u'Status': u'ACTIVE',
                u'simLaunch': u'2018-04-06T02:57:15.217442',
                u'simID': u'virlutils_default_XzmcAm',
                u'NodeName': u'router1',
                u'simStatus': u'ACTIVE',
                u'simExpires': None,
                u'managementIP': u'1.1.1.1',
                u'managementProtocol': u'telnet',
                u'SimulationHost': u'10.10.10.170',
                u'NodeSubtype': u'IOSv',
                u'PortMonitor': 17029,
                u'Reachable': True,
                u'SerialPorts': 2,
                u'managementProxy': u'jumphost',
                u'VncConsole': False,
                u'PortConsole': 17028,
                u'Annotation': u'REACHABLE'
            },
            u'guest|TEST_ENV|virl|router2': {
                u'Status': u'ACTIVE',
                u'simLaunch': u'2018-04-06T02:57:15.217442',
                u'simID': u'virlutils_default_XzmcAm',
                u'NodeName': u'router2',
                u'simStatus': u'ACTIVE',
                u'simExpires': None,
                u'managementIP': u'172.16.30.122',
                u'managementProtocol': u'telnet',
                u'SimulationHost': u'10.10.10.170',
                u'NodeSubtype': u'CSR1000',
                u'PortMonitor': 17030,
                u'Reachable': True,
                u'SerialPorts': 2,
                u'managementProxy': u'jumphost',
                u'VncConsole': False,
                u'PortConsole': 17027,
                u'Annotation': u'REACHABLE'
            },
            u'guest|TEST_ENV|virl|~mgmt-lxc': {
                u'Status': u'ACTIVE',
                u'simLaunch': u'2018-04-06T02:57:15.217442',
                u'simID': u'virlutils_default_XzmcAm',
                u'NodeName': u'router2',
                u'simStatus': u'ACTIVE',
                u'simExpires': None,
                u'managementIP': u'LXC_IP',
                u'externalAddr': u'FAKE_EXTERNAL_IP',
                u'managementProtocol': u'telnet',
                u'SimulationHost': u'10.10.10.170',
                u'NodeSubtype': u'mgmt-lxc',
                u'PortMonitor': 17030,
                u'Reachable': True,
                u'SerialPorts': 2,
                u'managementProxy': u'jumphost',
                u'VncConsole': False,
                u'PortConsole': 17027,
                u'Annotation': u'REACHABLE'
            },
            u'guest|TEST_ENV|virl|xr2': {
                u'Status': u'ACTIVE',
                u'simLaunch': u'2018-04-06T02:57:15.217442',
                u'simID': u'virlutils_default_XzmcAm',
                u'NodeName': u'router2',
                u'simStatus': u'ACTIVE',
                u'simExpires': None,
                u'managementIP': u'172.16.30.122',
                u'managementProtocol': u'telnet',
                u'SimulationHost': u'10.10.10.170',
                u'NodeSubtype': u'XR',
                u'PortMonitor': 17030,
                u'Reachable': True,
                u'SerialPorts': 2,
                u'managementProxy': u'jumphost',
                u'VncConsole': False,
                u'PortConsole': 17027,
                u'Annotation': u'REACHABLE'
            },
            u'guest|TEST_ENV|virl|nxos': {
                u'Status': u'ACTIVE',
                u'simLaunch': u'2018-04-06T02:57:15.217442',
                u'simID': u'virlutils_default_XzmcAm',
                u'NodeName': u'router2',
                u'simStatus': u'ACTIVE',
                u'simExpires': None,
                u'managementIP': u'172.16.30.122',
                u'managementProtocol': u'telnet',
                u'SimulationHost': u'10.10.10.170',
                u'NodeSubtype': u'NXOS',
                u'PortMonitor': 17030,
                u'Reachable': True,
                u'SerialPorts': 2,
                u'managementProxy': u'jumphost',
                u'VncConsole': False,
                u'PortConsole': 17027,
                u'Annotation': u'REACHABLE'
            },
            u'guest|TEST_ENV|virl|via-lxc': {
                u'Status': u'ACTIVE',
                u'simLaunch': u'2018-04-06T02:57:15.217442',
                u'simID': u'virlutils_default_XzmcAm',
                u'NodeName': u'router2',
                u'simStatus': u'ACTIVE',
                u'simExpires': None,
                u'managementIP': u'10.5.5.5',
                u'managementProtocol': u'telnet',
                u'SimulationHost': u'10.10.10.170',
                u'NodeSubtype': u'badsubtype',
                u'PortMonitor': 17030,
                u'Reachable': True,
                u'SerialPorts': 2,
                u'managementProxy': u'lxc',
                u'VncConsole': False,
                u'PortConsole': 17027,
                u'Annotation': u'REACHABLE'
            },
            u'guest|TEST_ENV|virl|badtype': {
                u'Status': u'ACTIVE',
                u'simLaunch': u'2018-04-06T02:57:15.217442',
                u'simID': u'virlutils_default_XzmcAm',
                u'NodeName': u'router2',
                u'simStatus': u'ACTIVE',
                u'simExpires': None,
                u'managementIP': u'172.16.30.122',
                u'managementProtocol': u'telnet',
                u'SimulationHost': u'10.10.10.170',
                u'NodeSubtype': u'badsubtype',
                u'PortMonitor': 17030,
                u'Reachable': True,
                u'SerialPorts': 2,
                u'managementProxy': u'jumphost',
                u'VncConsole': False,
                u'PortConsole': 17027,
                u'Annotation': u'REACHABLE'
            },
            u'guest|TEST_ENV|virl|keyerror': {
                u'Status': u'ACTIVE',
                u'PortMonitor': 17030,
                u'Reachable': True,
                u'SerialPorts': 2,
                u'managementProxy': u'jumphost',
                u'VncConsole': False,
                u'PortConsole': 17027,
                u'Annotation': u'REACHABLE'
            }
        }
        return response

    @classmethod
    def get_interfaces(cls):

        response = {
            "2nodes_default_ngTj03": {
                "router2": {
                    "0": {
                        "status": "UP",
                        "link-state": True,
                        "name": "GigabitEthernet0/1",
                        "hw-addr": "fa:16:3e:48:8e:c2",
                        "port-state": True,
                        "external-ip-address": None,
                        "ip-address": "10.255.0.6/30",
                        "external-port": None,
                        "network": "router1-to-router2"
                    },
                    "management": {
                        "status": "UP",
                        "link-state": True,
                        "name": "GigabitEthernet0/0",
                        "hw-addr": "5e:00:c0:01:00:00",
                        "port-state": True,
                        "external-ip-address": True,
                        "ip-address": "172.16.30.126/24",
                        "external-port": True,
                        "network": "flat"
                    }
                },
                "router1": {
                    "0": {
                        "status": "UP",
                        "link-state": True,
                        "name": "GigabitEthernet0/1",
                        "hw-addr": "fa:16:3e:95:f1:c5",
                        "port-state": True,
                        "external-ip-address": None,
                        "ip-address": "10.255.0.5/30",
                        "external-port": None,
                        "network": "router1-to-router2"
                    },
                    "management": {
                        "status": "UP",
                        "link-state": True,
                        "name": "GigabitEthernet0/0",
                        "hw-addr": "5e:00:c0:00:00:00",
                        "port-state": True,
                        "external-ip-address": None,
                        "ip-address": "172.16.30.125/24",
                        "external-port": None,
                        "network": "flat"
                    }
                }
            }
        }
        return response

    @classmethod
    def export(cls):

        with open('tests/v1/static/test_virl_data', 'r') as fh:
            virl_data = fh.read()
        return virl_data

    @classmethod
    def list_simulations(cls):
        response = {
            'simulations': {
                'sim1': {
                    'status': 'ACTIVE',
                    'expires': None,
                    'launched': '2017-12-08T23:39:07.721310',
                },
                'sim3': {
                    'status': 'BUILDING',
                    'expires': None,
                    'launched': '2017-12-08T18:48:34.174486',
                },
                'sim4': {
                    'status': 'ABSENT',
                    'expires': None,
                    'launched': '2017-12-08T18:48:34.174486',
                },
                'sim2': {
                    'status': 'ACTIVE',
                    'expires': None,
                    'launched': '2017-12-08T18:48:34.174486',
                },
                'virlutils-id': {
                    'status': 'ACTIVE',
                    'expires': None,
                    'launched': '2017-12-08T18:48:34.174486',
                },

            }
        }
        return response



import click
import tabulate


def node_def_list_table(image_list):

    headers = ["ID", "Label", "Description", "Max No. Interfaces", "RAM", "CPUs", "Boot Disk Size"]
    table = list()

    for f in list(image_list):
        tr = list()
        tr.append(str(f["id"]))
        tr.append(str(f["data"]["ui"].get("label", "N/A")))
        tr.append(str(f["data"]["general"]["description"]))
        tr.append(str(len(f["data"]["device"]["interfaces"]["physical"])))
        linux_native = f["data"]["sim"].get("linux_native", None)
        if linux_native:
            ram = int(linux_native.get("ram", 0))
            unit = "GB"
            if ram > 1024:
                ram /= 1024
            else:
                unit = "MB"
            tr.append(str(ram) + " " + unit)
            tr.append(str(linux_native.get("cpus", "N/A")))
            if "boot_disk_size" in linux_native:
                tr.append(str(linux_native["boot_disk_size"]) + " GB")
            else:
                tr.append("N/A")
        else:
            tr.append("N/A")  # RAM
            tr.append("N/A")  # CPU
            tr.append("N/A")  # Boot Disk Size

        table.append(tr)

    try:
        click.echo(tabulate.tabulate(table, headers, tablefmt="fancy_grid"))
    except UnicodeEncodeError:
        click.echo(tabulate.tabulate(table, headers, tablefmt="grid"))



from . import BaseTest
from click.testing import CliRunner
import requests_mock
from virl.cli.main import virl
try:
    from unittest.mock import patch
except ImportError:
    from mock import patch


class Test(BaseTest):

    def test_virl_logs(self):
        with requests_mock.mock() as m:
            # Mock the request to return what we expect from the API.
            m.get('http://localhost:19399/simengine/rest/events/TEST_ENV',
                  json=self.mock_request())
            runner = CliRunner()
            result = runner.invoke(virl, ["logs"])
            self.assertEqual(0, result.exit_code)

    @patch('virl.helpers.check_sim_running', return_value=False)
    def test_virl_logs_not_running(self, mock):

        with requests_mock.mock() as m:
            # Mock the request to return what we expect from the API.
            m.get('http://localhost:19399/simengine/rest/events/TEST_ENV',
                  json=self.mock_request())
            runner = CliRunner()
            result = runner.invoke(virl, ["logs"])
            self.assertEqual(0, result.exit_code)

    def mock_request(self):
        response = {
            u'events': [
                {
                    u'name': u'unspecified',
                    u'level': u'INFO',
                    u'session': u'virlutils_default_WPJ4ES',
                    u'user': u'guest',
                    u'time': u'Apr/04/201814:25:13',
                    u'message': u'some message',
                    u'id': 1457
                }
            ]
        }
        return response



import click
from virl.api import VIRLServer
from virl.helpers import get_cml_client


@click.command()
@click.argument("node", nargs=1)
@click.option("-f", "--filename", required=False, metavar="<filename>", help="filename to save to")
def export(node, filename):
    """
    export a node definition
    """

    server = VIRLServer()
    client = get_cml_client(server)

    if not filename:
        filename = node + ".yaml"

    defs = client.definitions

    try:
        ndef = defs.download_node_definition(node)
    except Exception as e:
        click.secho("Failed to download node definition for {}: {}".format(node, e), fg="red")
        exit(1)
    else:
        with open(filename, "w") as fd:
            fd.write(ndef)



from . import BaseTest
from click.testing import CliRunner
import requests_mock
from virl.cli.main import virl


class Tests(BaseTest):

    def test_virl_save(self):
        with requests_mock.mock() as m:
            # Mock the request to return what we expect from the API.
            down_url = 'http://localhost:19399/simengine/rest/stop/TEST_ENV'
            m.get(down_url, json=self.mock_down_response())
            runner = CliRunner()
            runner.invoke(virl, ["save"])

    def mock_down_response(self):
        response = 'SUCCESS'
        return response



import os
from . import BaseCMLTest
from click.testing import CliRunner
import requests_mock

try:
    from unittest.mock import patch
except ImportError:
    from mock import patch  # noqa


class CMLUseTest(BaseCMLTest):
    @patch("virl.cli.use.commands.call", auto_spec=False, return_value=0)
    def test_cml_use(self, call_mock):
        with requests_mock.Mocker() as m:
            self.setup_mocks(m)
            virl = self.get_virl()
            runner = CliRunner()
            runner.invoke(virl, ["use"])
            call_mock.assert_called_once_with(["virl", "use", "--help"])

    def test_cml_use_with_lab(self):
        with requests_mock.Mocker() as m:
            self.setup_mocks(m)
            virl = self.get_virl()
            runner = CliRunner()
            result = runner.invoke(virl, ["use", self.get_test_title()])
            self.assertEqual(0, result.exit_code)

    def test_cml_use_with_id(self):
        with requests_mock.Mocker() as m:
            self.setup_mocks(m)
            virl = self.get_virl()
            runner = CliRunner()
            result = runner.invoke(virl, ["use", "--id", self.get_test_id()])
            self.assertEqual(0, result.exit_code)

    def test_cml_use_with_lab_name(self):
        with requests_mock.Mocker() as m:
            self.setup_mocks(m)
            virl = self.get_virl()
            runner = CliRunner()
            result = runner.invoke(virl, ["use", "--lab-name", self.get_test_title()])
            self.assertEqual(0, result.exit_code)

    def test_cml_use_with_cache(self):
        try:
            os.remove(".virl/current_cml_lab")
        except OSError:
            pass

        try:
            os.remove(".virl/cached_cml_labs/{}".format(self.get_test_id()))
        except OSError:
            pass

        with requests_mock.Mocker() as m:
            self.setup_mocks(m)
            virl = self.get_virl()
            runner = CliRunner()
            result = runner.invoke(virl, ["use", "--id", self.get_test_id()])
            self.assertEqual(0, result.exit_code)

    def test_cml_use_with_bogus_id(self):
        with requests_mock.Mocker() as m:
            self.setup_mocks(m)
            virl = self.get_virl()
            runner = CliRunner()
            result = runner.invoke(virl, ["use", "--id", "123456"])
            self.assertEqual(1, result.exit_code)
            self.assertIn("Unable to find lab in the cache or on the server", result.output)



import click
import os
from virl.api import VIRLServer
from virl.helpers import get_cml_client


@click.command()
@click.argument("node", nargs=1)
@click.option("-f", "--filename", required=True, metavar="<filename>", help="path to the local node definition file")
def nimport(node, filename):
    """
    import a node definition
    """

    server = VIRLServer()
    client = get_cml_client(server)

    if not os.path.isfile(filename):
        click.secho("Node definition file {} does not exist or is not a file", fg="red")
        exit(1)
    else:
        defs = client.definitions
        contents = None

        with open(filename, "r") as fd:
            contents = fd.read()

        try:
            defs.upload_node_definition(node, contents)
        except Exception as e:
            click.secho("Failed to import node definition for {}: {}".format(node, e), fg="red")
            exit(1)



from . import BaseTest
from click.testing import CliRunner
import requests_mock
from virl.cli.main import virl


class FlavorsTest(BaseTest):

    def get_flavors(self):
        response = {
            "flavors": [
                {
                    "OS-FLV-DISABLED:disabled": False,
                    "OS-FLV-EXT-DATA:ephemeral": 0,
                    "disk": 0,
                    "id": "7ca78ccd-4a59-4b0c-802e-2a82de05e7c9",
                    "name": "ASAv",
                    "os-flavor-access:is_public": True,
                    "ram": 2048,
                    "rxtx_factor": 1.0,
                    "swap": "",
                    "vcpus": 1
                },
                {
                    "OS-FLV-DISABLED:disabled": False,
                    "OS-FLV-EXT-DATA:ephemeral": 0,
                    "disk": 0,
                    "id": "99710bf7-b822-4240-9d5e-17bc03393675",
                    "name": "CSR1000v",
                    "os-flavor-access:is_public": True,
                    "ram": 3072,
                    "rxtx_factor": 1.0,
                    "swap": "",
                    "vcpus": 1
                },
                {
                    "OS-FLV-DISABLED:disabled": False,
                    "OS-FLV-EXT-DATA:ephemeral": 0,
                    "disk": 0,
                    "id": "a484b396-c0fd-4949-80fb-7a0ac4e15a2f",
                    "name": "IOS XRv",
                    "os-flavor-access:is_public": True,
                    "ram": 3072,
                    "rxtx_factor": 1.0,
                    "swap": "",
                    "vcpus": 1
                },
                {
                    "OS-FLV-DISABLED:disabled": False,
                    "OS-FLV-EXT-DATA:ephemeral": 0,
                    "disk": 0,
                    "id": "a484b396-c0fd-4949-80fb-7a0ac4efffff",
                    "name": "mockTest",
                    "os-flavor-access:is_public": True,
                    "ram": 2048,
                    "rxtx_factor": 1.0,
                    "swap": "",
                    "vcpus": 1
                },
                {
                    "OS-FLV-DISABLED:disabled": False,
                    "OS-FLV-EXT-DATA:ephemeral": 0,
                    "disk": 0,
                    "id": "a484b396-c0fd-4949-80fb-7a0ac4eeeeee",
                    "name": "mockTest2",
                    "os-flavor-access:is_public": True,
                    "ram": 4096,
                    "rxtx_factor": 1.0,
                    "swap": "",
                    "vcpus": 2
                }
            ]
        }

        return response

    def return_flavor(self, name=None):
        if name is None:
            raise IndexError("No such index {}".format(name))

        r = self.get_flavors()
        if name == 'mockTest':
            return {"flavor": r['flavors'][3]}

        if name == 'mockTest2':
            return {"flavor": r['flavors'][4]}

    def test_virl_flavors_00_ls(self):
        with requests_mock.mock() as m:
            # Mock the request to return what we expect from the API.
            m.get('http://localhost:80/rest/flavors',
                  json=self.get_flavors())

            runner = CliRunner()
            result = runner.invoke(virl, ["flavors", "ls"])
            self.assertEqual(0, result.exit_code)

    def test_virl_flavors_01_add(self):
        with requests_mock.mock() as m:
            # Mock the request to return what we expect from the API.
            m.post('http://localhost:80/rest/flavors',
                   json=self.return_flavor('mockTest'))

            runner = CliRunner()
            result = runner.invoke(virl, ["flavors", "add", "mockTest"])
            self.assertEqual(0, result.exit_code)

            args = [
                "flavors", "add", "mockTest2",
                "--vcpus", "2",
                "--memory", "4096"
            ]
            result = runner.invoke(virl, args)
            self.assertEqual(0, result.exit_code)

    def test_virl_flavors_02_update(self):
        with requests_mock.mock() as m:
            # Mock the request to return what we expect from the API.
            m.get('http://localhost:80/rest/flavors',
                  json=self.get_flavors())
            m.post('http://localhost:80/rest/flavors',
                   json=self.return_flavor('mockTest'))

            u = 'http://localhost:80/rest/flavors/{}'
            u = u.format('a484b396-c0fd-4949-80fb-7a0ac4efffff')
            m.get(u, json=self.return_flavor('mockTest'))
            m.delete(u, json=self.return_flavor('mockTest'))

            runner = CliRunner()
            args = [
                "flavors", "update", "mockTest",
                "--vcpus", "2",
            ]
            result = runner.invoke(virl, args)
            self.assertEqual(0, result.exit_code)

    def test_virl_flavors_03_delete(self):
        with requests_mock.mock() as m:
            # Mock the request to return what we expect from the API.
            m.get('http://localhost:80/rest/flavors',
                  json=self.get_flavors())

            u = 'http://localhost:80/rest/flavors/{}'
            u = u.format('a484b396-c0fd-4949-80fb-7a0ac4efffff')
            m.delete(u, json=self.return_flavor('mockTest'))

            u = 'http://localhost:80/rest/flavors/{}'
            u = u.format('a484b396-c0fd-4949-80fb-7a0ac4eeeeee')
            m.delete(u, json=self.return_flavor('mockTest2'))

            runner = CliRunner()
            result = runner.invoke(virl, ["flavors", "delete", "mockTest"])
            self.assertEqual(0, result.exit_code)

            result = runner.invoke(virl, ["flavors", "delete", "mockTest2"])
            self.assertEqual(0, result.exit_code)



import click
from virl.api import VIRLServer
from requests.exceptions import HTTPError


@click.command()
@click.argument('flavor')
@click.option('--memory', default='2048')
@click.option('--vcpus', default='1')
def add(**kwargs):
    """
    add a specific flavor
    """

    flavor = kwargs.get('flavor')
    memory = kwargs.get('memory')
    vcpus = kwargs.get('vcpus')
    server = VIRLServer(port=80)

    # Attempt to delete the flavor
    try:
        r = server.add_flavor(flavor=flavor, memory=memory, vcpus=vcpus)
        print("Flavor '{}' ({}) added.".format(flavor, r['id']))

    except HTTPError as err:
        print("\nFailed to create flavor '{}':".format(flavor))
        print("\n\t{}\n".format(err.response.reason))



# -*- coding: utf-8 -*-
import click
import tabulate


def lab_list_table(labs, cached_labs=None):
    click.secho("Labs on Server", fg="green")
    print_labs(labs)
    if cached_labs:
        click.secho("Cached Labs", fg="yellow")
        print_labs(cached_labs)


def print_labs(labs):
    table = list()
    # TODO: Do we truncate description as it can be kind of long?
    headers = ["ID", "Title", "Description", "Owner", "Status", "Nodes", "Links", "Interfaces"]
    for lab in labs:
        tr = list()
        tr.append(lab.id)
        tr.append(lab.title)
        tr.append(lab.description)
        tr.append(lab.owner)
        status = lab.state()
        stats = lab.statistics
        if status in {"BOOTED", "STARTED"}:
            color = "green"
        elif status in {"QUEUED"}:
            color = "yellow"
        else:
            color = "red"
        tr.append(click.style(status, fg=color))
        tr.append(stats["nodes"])
        tr.append(stats["links"])
        tr.append(stats["interfaces"])
        table.append(tr)
    # wrap the output in this try/except block as some terminals
    # may have problem with the 'fancy_grid'
    try:
        click.echo(tabulate.tabulate(table, headers, tablefmt="fancy_grid"))
    except UnicodeEncodeError:
        click.echo(tabulate.tabulate(table, headers, tablefmt="grid"))



# coding: utf-8
from setuptools import setup, find_packages  # noqa: H301
from virl import __version__

NAME = "virlutils"
CMLNAME = "cmlutils"
VERSION = __version__
# To install the library, run the following
#
# python setup.py install
#
# prerequisite: setuptools
# http://pypi.python.org/pypi/setuptools


def requirements(f):
    with open(f, "r") as fd:
        return fd.read()


def readme():
    with open("README.md", "r") as f:
        return f.read()


setup(
    name=NAME,
    version=VERSION,
    description="A collection of utilities for interacting with Cisco VIRL/CML",
    author="Joe Clarke",  # With a big thanks to its original author, Kevin Corbin
    author_email="jclarke@cisco.com",
    url="https://github.com/CiscoDevNet/virlutils",
    entry_points={"console_scripts": ["virl=virl.cli.main:virl", "cml=virl.cli.main:virl"]},
    packages=find_packages(),
    package_data={"virl": ["templates/**/*.j2", "swagger/templates/*", "swagger/static/*", "examples/plugins/*"]},
    include_package_data=True,
    install_requires=requirements("requirements.txt"),
    long_description_content_type="text/markdown",
    long_description=readme(),
    test_suite="tests",
    tests_require=requirements("test-requirements.txt"),
    zip_safe=False,
    python_requires=">=3.6",
    classifiers=[
        "Programming Language :: Python :: 3",
        "License :: OSI Approved :: MIT License",
        "Operating System :: OS Independent",
    ],
)



import click
from virl.api import VIRLServer
from virl import helpers
from virl.helpers import get_cml_client, get_current_lab, safe_join_existing_lab
from virl.generators import ansible_inventory_generator1, ansible_inventory_generator


@click.command()
@click.option("--output", "-o", help="output File name ")
@click.option("--style", help="output format (default is yaml)", type=click.Choice(["ini", "yaml"]))
def ansible(**kwargs):
    """
    generate ansible inventory
    """
    server = VIRLServer()
    client = get_cml_client(server)

    current_lab = get_current_lab()
    if current_lab:
        lab = safe_join_existing_lab(current_lab, client)
        if lab:

            if kwargs.get("output"):
                file_name = kwargs.get("output")
            elif kwargs.get("style") == "ini":
                file_name = "{}_inventory.ini".format(lab.id)
            else:
                file_name = "{}_inventory.yaml".format(lab.id)

            inv = None

            if kwargs.get("style") == "ini":
                inv = ansible_inventory_generator(lab, server, style="ini")
            else:
                inv = ansible_inventory_generator(lab, server)

            if inv:
                click.secho("Writing {}".format(file_name))
                with open(file_name, "w") as fd:
                    fd.write(inv)
            else:
                click.secho("Failed to get inventory data", fg="red")
                exit(1)
        else:
            click.secho("Failed to find running lab {}".format(current_lab), fg="red")
            exit(1)
    else:
        click.secho("Current lab is not set", fg="red")
        exit(1)


@click.command()
@click.argument("env", default="default")
@click.option("--output", "-o", help="output File name ")
@click.option("--style", help="output format (default is yaml)", type=click.Choice(["ini", "yaml"]))
def ansible1(env, **kwargs):
    """
    generate ansible inventory
    """

    if kwargs.get("output"):
        # user specified output filename
        file_name = kwargs.get("output")
    elif kwargs.get("style") == "ini":
        file_name = "{}_inventory.ini".format(env)
    else:
        # writes to <env>_testbed.yaml by default
        file_name = "{}_inventory.yaml".format(env)

    running = helpers.check_sim_running(env)
    if running:
        sim_name = running
        server = VIRLServer()
        roster = server.get_sim_roster(sim_name)
        # sim_name = "topology-fpyHFs"
        virl_data = server.export(sim_name, ip=True).content
        interfaces = server.get_interfaces(sim_name).json()
        if kwargs.get("style") == "ini":
            inventory_ini = ansible_inventory_generator1(sim_name, virl_data, roster=roster, interfaces=interfaces, style="ini")

            click.secho("Writing {}".format(file_name))
            with open(file_name, "w") as ini_file:
                ini_file.write(inventory_ini)
        else:
            inventory_yaml = ansible_inventory_generator1(sim_name, virl_data, roster=roster, interfaces=interfaces)

            click.secho("Writing {}".format(file_name))
            with open(file_name, "w") as yaml_file:
                yaml_file.write(inventory_yaml)

    else:
        click.secho("couldnt generate testbed for for env: {}".format(env), fg="red")



import click
import tabulate


def flavor_list_table(flavors_list):

    headers = ['Name', 'Memory', 'vCPUs']
    table = list()

    for f in list(flavors_list):
        tr = list()
        tr.append(str(f['name']))
        tr.append(str(f['ram']))
        tr.append(str(f['vcpus']))

        table.append(tr)

    try:
        click.echo(tabulate.tabulate(table, headers, tablefmt="fancy_grid"))
    except UnicodeEncodeError:
        click.echo(tabulate.tabulate(table, headers, tablefmt="grid"))



import click
import os
from virl.api import VIRLServer
from virl.helpers import get_cml_client, get_current_lab, safe_join_existing_lab
from virl2_client.models.cl_pyats import ClPyats, PyatsNotInstalled, PyatsDeviceNotFound


@click.command()
@click.argument("node", nargs=1)
@click.argument("command", nargs=1)
@click.option("--config/--no-config", default=False, show_default=False, help="Command is a configuration command (default: False)")
def command(node, command, config, **kwargs):
    """
    send a command or config to a node (requires pyATS)
    """
    server = VIRLServer()
    client = get_cml_client(server)

    current_lab = get_current_lab()
    if current_lab:
        lab = safe_join_existing_lab(current_lab, client)
        if lab:
            pylab = None
            try:
                pylab = ClPyats(lab)
            except PyatsNotInstalled:
                click.secho("pyATS is not installed, run 'pip install pyats'", fg="red")
                exit(1)

            pyats_username = server.config.get("CML_DEVICE_USERNAME")
            pyats_password = server.config.get("CML_DEVICE_PASSWORD")
            pyats_auth_password = server.config.get("CML_DEVICE_ENABLE_PASSWORD")

            if pyats_username:
                os.environ["PYATS_USERNAME"] = pyats_username
            if pyats_password:
                os.environ["PYATS_PASSWORD"] = pyats_password
            if pyats_auth_password:
                os.environ["PYATS_AUTH_PASS"] = pyats_auth_password

            pylab.sync_testbed(server.user, server.passwd)

            try:
                result = ""
                if config:
                    result = pylab.run_config_command(node, command)
                else:
                    result = pylab.run_command(node, command)

                click.secho(result)
            except PyatsDeviceNotFound:
                click.secho("Node '{}' is not supported by pyATS".format(node), fg="yellow")
            except Exception as e:
                click.secho("Failed to run '{}' on '{}': {}".format(command, node, e))
                exit(1)
        else:
            click.secho("Unable to find lab {}".format(current_lab), fg="red")
            exit(1)
    else:
        click.secho("No current lab set", fg="red")
        exit(1)



import click
from subprocess import call
from virl.api import VIRLServer
from virl.helpers import (
    store_sim_info,
    get_cml_client,
    safe_join_existing_lab_by_title,
    safe_join_existing_lab,
    check_lab_cache,
    cache_lab,
    set_current_lab,
    get_command,
)


# This may need to become a helper
def check_lab_cache_server(lab_id, client):
    """
    check if a lab exists in either the cache or the server.
    if on server and not in cache, cache the lab.
    """
    ret = None

    if not check_lab_cache(lab_id):
        lab_obj = safe_join_existing_lab(lab_id, client)
        if lab_obj:
            cache_lab(lab_obj)
            ret = lab_id
    else:
        ret = lab_id

    return ret


@click.command()
@click.argument("lab", required=False)
@click.option("--id", required=False, help="An existing lab ID to make the current lab (lab-name is ignored)")
@click.option("--lab-name", "-n", required=False, help="An existing lab name to make the current lab")
def use(lab, id, lab_name):
    """
    use lab launched elsewhere
    """
    server = VIRLServer()
    client = get_cml_client(server)
    lab_id = None

    if not lab and not id and not lab_name:
        exit(call([get_command(), "use", "--help"]))

    if id:
        lab_id = check_lab_cache_server(id, client)

    # Prefer --lab-name over positional argument
    if lab_name:
        lab = lab_name

    if not id and lab:
        lab_obj = safe_join_existing_lab_by_title(lab, client)
        if lab_obj:
            # Make sure this lab is cached.
            lab_id = check_lab_cache_server(lab_obj.id, client)

    if lab_id:
        set_current_lab(lab_id)
    else:
        click.secho("Unable to find lab in the cache or on the server", fg="red")
        exit(1)


@click.command()
@click.argument("sim")
def use1(sim):
    """
    use virl simulation launched elsewhere
    """
    store_sim_info(sim, env="default")  # 'topology-2lkx2'
    click.secho("Now using VIRL simulation {}".format(sim))
